# -*- coding: utf-8 -*-

import logging
import warnings

from collections import namedtuple, defaultdict

from typing import Any, DefaultDict, Dict, Generic, Optional, List, Set, Tuple, TypeVar, Union, Sequence

import numpy as np

from ms_deisotope.data_source import ProcessedScan
from ms_deisotope.peak_set import DeconvolutedPeak

from glypy.structure.glycan_composition import FrozenMonosaccharideResidue, HashableGlycanComposition, GlycanComposition
from glypy.utils.enum import EnumValue

from glycopeptidepy.structure.fragment import StubFragment
from glycopeptidepy.structure.fragmentation_strategy import StubGlycopeptideStrategy, _AccumulatorBag
from glycopeptidepy.structure.fragmentation_strategy.glycan import GlycanCompositionFragment


from glycresoft.chromatogram_tree.mass_shift import MassShift
from glycresoft.serialize import GlycanCombination, GlycanTypes
from glycresoft.database.disk_backed_database import PPMQueryInterval
from glycresoft.chromatogram_tree import Unmodified
from glycresoft.structure.denovo import PathSet, PathFinder, Path

logger = logging.getLogger("glycresoft.core_search")


hexnac = FrozenMonosaccharideResidue.from_iupac_lite("HexNAc")
hexose = FrozenMonosaccharideResidue.from_iupac_lite("Hex")
xylose = FrozenMonosaccharideResidue.from_iupac_lite("Xyl")
fucose = FrozenMonosaccharideResidue.from_iupac_lite("Fuc")
dhex = FrozenMonosaccharideResidue.from_iupac_lite("dHex")
neuac = FrozenMonosaccharideResidue.from_iupac_lite("NeuAc")
neugc = FrozenMonosaccharideResidue.from_iupac_lite("NeuGc")


def approximate_internal_size_of_glycan(glycan_composition):
    terminal_groups = glycan_composition._getitem_fast(neuac) +\
        glycan_composition._getitem_fast(neugc)
    side_groups = glycan_composition._getitem_fast(fucose) + glycan_composition._getitem_fast(dhex)
    n = sum(glycan_composition.values())
    n -= terminal_groups
    if side_groups > 1:
        n -= 1
    return n


def glycan_side_group_count(glycan_composition):
    side_groups = glycan_composition._getitem_fast(
        fucose) + glycan_composition._getitem_fast(dhex)
    return side_groups


def isclose(a, b, rtol=1e-05, atol=1e-08):
    return abs(a - b) <= atol + rtol * abs(b)


default_components = (hexnac, hexose, xylose, fucose,)


class CoreMotifFinder(PathFinder):
    minimum_peptide_mass: float

    def __init__(self, components=None, product_error_tolerance=1e-5, minimum_peptide_mass=350.):
        if components is None:
            components = default_components
        super().__init__(components, product_error_tolerance, False)
        self.minimum_peptide_mass = minimum_peptide_mass

    def find_n_linked_core(self, groups: DefaultDict[Any, List], min_size=1):
        sequence = [hexnac, hexnac, hexose, hexose, hexose]
        expected_n = len(sequence)
        terminals = dict()

        for label, paths in groups.items():
            label_i = 0
            expected_i = 0
            path_n = len(label)
            while label_i < path_n and expected_i < expected_n:
                edge = label[label_i]
                label_i += 1
                expected = sequence[expected_i]
                if expected == edge:
                    expected_i += 1
                elif edge == fucose or edge == dhex:
                    continue
                else:
                    break
            if expected_i >= min_size:
                for path in paths:
                    last_path = terminals.get(path[0].start)
                    if last_path is None:
                        terminals[path[0].start] = path
                    else:
                        terminals[path[0].start] = max((path, last_path), key=lambda x: x.total_signal)
        return PathSet(terminals.values())

    def find_o_linked_core(self, groups: DefaultDict[Any, List], min_size=1):
        sequence = [(hexnac, hexose), (hexnac, hexose, fucose,), (hexnac, hexose, fucose,)]
        expected_n = len(sequence)
        terminals = dict()

        for label, paths in groups.items():
            label_i = 0
            expected_i = 0
            path_n = len(label)
            while label_i < path_n and expected_i < expected_n:
                edge = label[label_i]
                label_i += 1
                expected = sequence[expected_i]
                if edge in expected:
                    expected_i += 1
                else:
                    break
            if expected_i >= min_size:
                for path in paths:
                    last_path = terminals.get(path[0].start)
                    if last_path is None:
                        terminals[path[0].start] = path
                    else:
                        terminals[path[0].start] = max((path, last_path), key=lambda x: x.total_signal)
        return PathSet(terminals.values())

    def find_gag_linker_core(self, groups: DefaultDict[Any, List], min_size=1):
        sequence = [xylose, hexose, hexose, ]
        expected_n = len(sequence)
        terminals = dict()

        for label, paths in groups.items():
            label_i = 0
            expected_i = 0
            path_n = len(label)
            while label_i < path_n and expected_i < expected_n:
                edge = label[label_i]
                label_i += 1
                expected = sequence[expected_i]
                if expected == edge:
                    expected_i += 1
                elif edge == fucose:
                    continue
                else:
                    break
            if expected_i >= min_size:
                for path in paths:
                    last_path = terminals.get(path[0].start)
                    if last_path is None:
                        terminals[path[0].start] = path
                    else:
                        terminals[path[0].start] = max((path, last_path), key=lambda x: x.total_signal)
        return PathSet(terminals.values())

    def estimate_peptide_mass(self, scan: ProcessedScan, topn: int=100, mass_shift: MassShift=Unmodified,
                              query_mass: Optional[float] = None,
                              simplify: bool = True) -> Union[List[float], List[Tuple[float, List[Path]]]]:
        graph = self.build_graph(scan, mass_shift=mass_shift)
        paths = self.paths_from_graph(graph)
        groups = self.aggregate_paths(paths)

        n_linked_paths = self.find_n_linked_core(groups)
        o_linked_paths = self.find_o_linked_core(groups)
        gag_linker_paths = self.find_gag_linker_core(groups)
        peptide_masses = []

        has_tandem_shift = abs(mass_shift.tandem_mass) > 0

        # TODO: split the different motif masses up according to core type efficiently
        # but for now just lump them all together
        for path in n_linked_paths:
            if path.start_mass < self.minimum_peptide_mass:
                continue
            peptide_masses.append((path.start_mass, path))
            if has_tandem_shift:
                peptide_masses.append((path.start_mass - mass_shift.tandem_mass, path))
        for path in o_linked_paths:
            if path.start_mass < self.minimum_peptide_mass:
                continue
            peptide_masses.append((path.start_mass, path))
            if has_tandem_shift:
                peptide_masses.append((path.start_mass - mass_shift.tandem_mass, path))
        for path in gag_linker_paths:
            if path.start_mass < self.minimum_peptide_mass:
                continue
            peptide_masses.append((path.start_mass, path))
            if has_tandem_shift:
                peptide_masses.append((path.start_mass - mass_shift.tandem_mass, path))
        peptide_masses.sort(key=lambda x: x[0])
        result = []
        paths_for = []
        last = 0
        for m, path in peptide_masses:
            if abs(last - m) < 1e-3:
                if not simplify:
                    paths_for.append(path)
                continue
            if simplify:
                result.append(m)
            else:
                paths_for = [path]
                result.append((m, paths_for))
            last = m
        return result[:topn]

    def build_peptide_filter(self, scan, error_tolerance=1e-5, mass_shift=Unmodified, query_mass=None):
        peptide_masses = self.estimate_peptide_mass(
            scan, mass_shift=mass_shift, query_mass=query_mass, simplify=True)

        out = []
        if len(peptide_masses) == 0:
            return IntervalFilter([])
        last = PPMQueryInterval(peptide_masses[0], error_tolerance)
        for point in peptide_masses[1:]:
            interval = PPMQueryInterval(point, error_tolerance)
            if interval.overlaps(last):
                last.extend(interval)
            else:
                out.append(last)
                last = interval
        out.append(last)
        return IntervalFilter(out)


class CoarseStubGlycopeptideFragment(object):
    __slots__ = ['key', 'is_core', 'mass']

    def __init__(self, key, mass, is_core):
        self.key = key
        self.mass = mass
        self.is_core = is_core

    def __eq__(self, other):
        try:
            return self.key == other.key and self.is_core == other.is_core
        except AttributeError:
            return self.key == other

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(int(self.mass))

    def __lt__(self, other):
        return self.mass < other.mass

    def __gt__(self, other):
        return self.mass > other.mass

    def __reduce__(self):
        return self.__class__, (self.key, self.mass, self.is_core)

    def __repr__(self):
        return "%s(%s, %f, %r)" % (
            self.__class__.__name__,
            self.key, self.mass, self.is_core
        )


class GlycanCombinationRecordBase(object):
    __slots__ = ['id', 'dehydrated_mass', 'composition', 'count', 'glycan_types',
                 'size', "_fragment_cache", "internal_size_approximation", "_hash",
                 'fragment_set_properties']

    id: int
    dehydrated_mass: float
    composition: GlycanComposition
    count: int
    glycan_types: List[EnumValue]
    size: int
    _fragment_cache: Dict[EnumValue, List]
    internal_size_approximation: int
    _hash: int
    fragment_set_properties: dict

    def is_n_glycan(self) -> bool:
        return GlycanTypes.n_glycan in self.glycan_types

    def is_o_glycan(self) -> bool:
        return GlycanTypes.o_glycan in self.glycan_types

    def is_gag_linker(self) -> bool:
        return GlycanTypes.gag_linker in self.glycan_types

    def get_n_glycan_fragments(self) -> List[CoarseStubGlycopeptideFragment]:
        if GlycanTypes.n_glycan not in self._fragment_cache:
            strategy = StubGlycopeptideStrategy(None, extended=True)
            shifts = strategy.n_glycan_composition_fragments(
                self.composition, 1, 0)
            fragment_structs = []
            for shift in shifts:
                if shift["key"]['HexNAc'] <= 2 and shift["key"]["Hex"] <= 3:
                    is_core = True
                else:
                    is_core = False
                fragment_structs.append(
                    CoarseStubGlycopeptideFragment(
                        shift['key'], shift['mass'], is_core))
            self._fragment_cache[GlycanTypes.n_glycan] = sorted(
                set(fragment_structs))
            return self._fragment_cache[GlycanTypes.n_glycan]
        else:
            return self._fragment_cache[GlycanTypes.n_glycan]

    def get_o_glycan_fragments(self) -> List[CoarseStubGlycopeptideFragment]:
        if GlycanTypes.o_glycan not in self._fragment_cache:
            strategy = StubGlycopeptideStrategy(None, extended=True)
            shifts = strategy.o_glycan_composition_fragments(
                self.composition, 1, 0)
            fragment_structs = []
            for shift in shifts:
                shift['key'] = _AccumulatorBag(shift['key'])
                fragment_structs.append(
                    CoarseStubGlycopeptideFragment(
                        shift['key'], shift['mass'], True))
            self._fragment_cache[GlycanTypes.o_glycan] = sorted(
                set(fragment_structs))
            return self._fragment_cache[GlycanTypes.o_glycan]
        else:
            return self._fragment_cache[GlycanTypes.o_glycan]

    def get_gag_linker_glycan_fragments(self) -> List[CoarseStubGlycopeptideFragment]:
        if GlycanTypes.gag_linker not in self._fragment_cache:
            strategy = StubGlycopeptideStrategy(None, extended=True)
            shifts = strategy.gag_linker_composition_fragments(
                self.composition, 1, 0)
            fragment_structs = []
            for shift in shifts:
                shift['key'] = _AccumulatorBag(shift['key'])
                fragment_structs.append(
                    CoarseStubGlycopeptideFragment(
                        shift['key'], shift['mass'], True))
            self._fragment_cache[GlycanTypes.gag_linker] = sorted(
                set(fragment_structs))
            return self._fragment_cache[GlycanTypes.gag_linker]
        else:
            return self._fragment_cache[GlycanTypes.gag_linker]

    def clear(self):
        self._fragment_cache.clear()

try:
    _GlycanCombinationRecordBase = GlycanCombinationRecordBase
    from glycresoft._c.tandem.core_search import GlycanCombinationRecordBase
except ImportError as err:
    print(err)
    pass


class GlycanCombinationRecord(GlycanCombinationRecordBase):
    """Represent a glycan combination compactly in memory

    Attributes
    ----------
    composition : :class:`~.HashableGlycanComposition`
        The glycan combination's composition in monosaccharide units
    count : int
        The number of distinct glycans this combination contains
    dehydrated_mass : float
        The total mass shift applied to a peptide when this combination is attached
        to it
    glycan_types : list
        The types of glycans combined to make this entity
    """

    __slots__ = ()

    @classmethod
    def from_combination(cls, combination):
        inst = cls(
            id=combination.id,
            dehydrated_mass=combination.dehydrated_mass(),
            composition=combination.convert(),
            count=combination.count,
            glycan_types=tuple(set([
                c.name for component_classes in combination.component_classes
                for c in component_classes])),
        )
        return inst

    @classmethod
    def from_hypothesis(cls, session, hypothesis_id):
        query = session.query(GlycanCombination).filter(
            GlycanCombination.hypothesis_id == hypothesis_id).group_by(
                GlycanCombination.composition, GlycanCombination.count).order_by(
                    GlycanCombination.dehydrated_mass()) # pylint: disable=no-value-for-parameter
        candidates = query.all()
        out = []
        for candidate in candidates:
            out.append(cls.from_combination(candidate))
        return out

    def _to_dict(self):
        return {
            "id": self.id,
            "dehydrated_mass": self.dehydrated_mass,
            "composition": str(self.composition),
            "count": self.count,
            "glycan_types": list(map(str, self.glycan_types)),
        }

    @classmethod
    def _from_dict(cls, d):
        d['composition'] = HashableGlycanComposition.parse(d['composition'])
        d['glycan_types'] = [GlycanTypes[t] for t in d['glycan_types']]
        return cls(**d)

    def __init__(self, id, dehydrated_mass, composition, count, glycan_types):
        self.id = id
        self.dehydrated_mass = dehydrated_mass
        self.composition = composition
        self.size = sum(composition.values())
        self.internal_size_approximation = self._approximate_total_size()
        self.side_group_count = glycan_side_group_count(self.composition)
        self.count = count
        self.glycan_types = list(glycan_types)
        self._fragment_cache = dict()
        self._hash = hash(self.composition)
        self.fragment_set_properties = dict()

    def __eq__(self, other):
        return (self.composition == other.composition) and (self.count == other.count) and (
            self.glycan_types == other.glycan_types)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return self._hash

    def _approximate_total_size(self):
        return approximate_internal_size_of_glycan(self.composition)

    def __reduce__(self):
        return (
            GlycanCombinationRecord,
            (self.id, self.dehydrated_mass, self.composition, self.count, self.glycan_types),
            self.__getstate__()
        )

    def __getstate__(self):
        return {
            "fragment_set_properties": self.fragment_set_properties
        }

    def __setstate__(self, state):
        self.fragment_set_properties = state['fragment_set_properties']

    def __repr__(self):
        return "GlycanCombinationRecord(%s, %d)" % (self.composition, self.count)


class CoarseStubGlycopeptideMatch(object):
    def __init__(self, key, mass, shift_mass, peaks_matched):
        self.key = key
        self.mass = mass
        self.shift_mass = shift_mass
        self.peaks_matched = peaks_matched

    def __reduce__(self):
        return self.__class__, (self.key, self.mass, self.shift_mass, self.peaks_matched)

    def __repr__(self):
        return "%s(%s, %f, %f, %r)" % (
            self.__class__.__name__,
            self.key, self.mass, self.shift_mass, self.peaks_matched
        )


class CoarseGlycanMatch(object):
    def __init__(self, matched_fragments, n_matched, n_theoretical, core_matched, core_theoretical):
        self.fragment_matches = list(matched_fragments)
        self.n_matched = n_matched
        self.n_theoretical = n_theoretical
        self.core_matched = core_matched
        self.core_theoretical = core_theoretical

    def __iter__(self):
        yield self.matched_fragments
        yield self.n_matched
        yield self.n_theoretical
        yield self.core_matched
        yield self.core_theoretical

    def estimate_peptide_mass(self):
        weighted_mass_acc = 0.0
        weight_acc = 0.0

        for fmatch in self.fragment_matches:
            fmass = fmatch.shift_mass
            for peak in fmatch.peaks_matched:
                weighted_mass_acc += (peak.neutral_mass - fmass) * peak.intensity
                weight_acc += peak.intensity
        if weight_acc == 0:
            return -1
        return weighted_mass_acc / weight_acc

    def __repr__(self):
        template = (
            "{self.__class__.__name__}({self.n_matched}, {self.n_theoretical}, "
            "{self.core_matched}, {self.core_theoretical})")
        return template.format(self=self)


class GlycanCoarseScorerBase(object):
    product_error_tolerance: float
    fragment_weight: float
    core_weight: float

    def __init__(self, product_error_tolerance=1e-5, fragment_weight=0.56, core_weight=0.42):
        self.product_error_tolerance = product_error_tolerance
        self.fragment_weight = fragment_weight
        self.core_weight = core_weight

    def _match_fragments(self, scan, peptide_mass, shifts, mass_shift_tandem_mass=0.0):
        fragment_matches = []
        core_matched = 0.0
        core_theoretical = 0.0
        has_tandem_shift = abs(mass_shift_tandem_mass) > 0
        for shift in shifts:
            if shift.is_core:
                is_core = True
                core_theoretical += 1
            else:
                is_core = False
            target_mass = shift.mass + peptide_mass
            hits = scan.deconvoluted_peak_set.all_peaks_for(target_mass, self.product_error_tolerance)
            if hits:
                if is_core:
                    core_matched += 1
                fragment_matches.append((shift.key, target_mass, hits))
            if has_tandem_shift:
                shifted_mass = target_mass + mass_shift_tandem_mass
                hits = scan.deconvoluted_peak_set.all_peaks_for(
                    shifted_mass, self.product_error_tolerance)
                if hits:
                    if is_core:
                        core_matched += 1
                    fragment_matches.append(
                        CoarseStubGlycopeptideMatch(
                            shift.key, shifted_mass, shift.mass + mass_shift_tandem_mass, hits))

        return CoarseGlycanMatch(
            fragment_matches, float(len(fragment_matches)), float(len(shifts)), core_matched, core_theoretical)

    # consider adding the internal size approximation to this method and it's Cython implementation.
    def _calculate_score(self, glycan_match):
        ratio_fragments = (glycan_match.n_matched / glycan_match.n_theoretical)
        ratio_core = glycan_match.core_matched / glycan_match.core_theoretical
        coverage = (ratio_fragments ** self.fragment_weight) * (ratio_core ** self.core_weight)
        score = 0
        for fmatch in glycan_match.fragment_matches:
            mass = fmatch.mass
            for peak in fmatch.matches:
                score += np.log(peak.intensity) * (1 - (np.abs(peak.neutral_mass - mass) / mass) ** 4) * coverage
        return score

    def _n_glycan_match_stubs(self, scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=0.0):
        shifts = glycan_combination.get_n_glycan_fragments()
        return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)

    def _o_glycan_match_stubs(self, scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=0.0):
        shifts = glycan_combination.get_o_glycan_fragments()
        return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)

    def _gag_match_stubs(self, scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=0.0):
        shifts = glycan_combination.get_gag_linker_glycan_fragments()
        return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)

try:
    _GlycanCoarseScorerBase = GlycanCoarseScorerBase
    from glycresoft._c.tandem.core_search import GlycanCoarseScorerBase
except ImportError:
    pass


_GlycanMatchResult = namedtuple('GlycanMatchResult', (
    'peptide_mass', 'score', 'match', 'glycan_size', 'glycan_types',
    'recalibrated_peptide_mass'))


class GlycanMatchResult(_GlycanMatchResult):
    __slots__ = ()

    @property
    def fragment_match_count(self):
        match = self.match
        if match is None:
            return 0
        return match.n_matched


def group_by_score(matches, threshold=1e-2):
    matches = sorted(matches, key=lambda x: x.score, reverse=True)
    groups = []
    if len(matches) == 0:
        return groups
    current_group = [matches[0]]
    last_match = matches[0]
    for match in matches[1:]:
        delta = abs(match.score - last_match.score)
        if delta > threshold:
            groups.append(current_group)
            current_group = [match]
        else:
            current_group.append(match)
        last_match = match
    groups.append(current_group)
    return groups


def flatten(groups):
    return [b for a in groups for b in a]


SolutionType = TypeVar("SolutionType")


class GlycanFilteringPeptideMassEstimatorBase(Generic[SolutionType]):
    use_denovo_motif: bool
    motif_finder: CoreMotifFinder
    glycan_combination_db: List[GlycanCombinationRecord]
    minimum_peptide_mass: float
    use_recalibrated_peptide_mass: float

    def __init__(self, glycan_combination_db, product_error_tolerance=1e-5,
                 fragment_weight=0.56, core_weight=0.42, minimum_peptide_mass=500.0,
                 use_denovo_motif=False, components=None,
                 use_recalibrated_peptide_mass=False):
        if not isinstance(glycan_combination_db[0], GlycanCombinationRecord):
            glycan_combination_db = [GlycanCombinationRecord.from_combination(gc)
                                     for gc in glycan_combination_db]
        self.use_denovo_motif = use_denovo_motif
        self.motif_finder = CoreMotifFinder(
            components, product_error_tolerance)
        self.glycan_combination_db = sorted(
            glycan_combination_db, key=lambda x: (x.dehydrated_mass, x.id))
        self.minimum_peptide_mass = minimum_peptide_mass
        self.use_recalibrated_peptide_mass = use_recalibrated_peptide_mass
        super(GlycanFilteringPeptideMassEstimatorBase, self).__init__(
            product_error_tolerance, fragment_weight, core_weight)

    def match(self, scan: ProcessedScan, mass_shift=Unmodified, query_mass=None) -> List[SolutionType]:
        raise NotImplementedError()

    def estimate_peptide_mass(self, scan: ProcessedScan,
                              topn: int=150,
                              threshold: float=-1,
                              min_fragments: int=0,
                              mass_shift: MassShift=Unmodified,
                              simplify: bool=True,
                              query_mass: Optional[float] = None) -> Union[List[float], List[SolutionType]]:
        '''Given an scan, estimate the possible peptide masses using the connected glycan database and
        mass differences from the precursor mass.

        Parameters
        ----------
        scan : ProcessedScan
            The deconvoluted scan to search
        topn : int, optional
            The number of solutions to return, sorted by quality descending
        threshold : float, optional
            The minimum match score to allow a returned solution to have.
        min_fragments : int, optional
            The minimum number of matched fragments to require a solution to have, independent
            of score.
        mass_shift : MassShift, optional
            The mass shift to apply to
        '''
        out = self.match(scan, mass_shift=mass_shift, query_mass=query_mass)
        out = [x for x in out if x.score >
               threshold and x.fragment_match_count >= min_fragments]
        groups = group_by_score(out)
        out = flatten(groups[:topn])
        if simplify:
            return [x.peptide_mass for x in out]
        return out

    def glycan_for_peptide_mass(self, scan: ProcessedScan, peptide_mass: float) -> List[GlycanCombinationRecord]:
        matches = []
        try:
            glycan_mass = scan.precursor_information.neutral_mass - peptide_mass
        except AttributeError:
            glycan_mass = scan - peptide_mass
        for glycan_record in self.glycan_combination_db:
            if abs(glycan_record.dehydrated_mass - glycan_mass) / glycan_mass < self.product_error_tolerance:
                matches.append(glycan_record)
            elif glycan_mass > glycan_record.dehydrated_mass:
                break
        return matches

    def build_peptide_filter(self, scan: ProcessedScan,
                             error_tolerance: Optional[float] = None,
                             mass_shift: MassShift = Unmodified,
                             query_mass: Optional[float] = None) -> 'IntervalFilter':
        if error_tolerance is None:
            error_tolerance = self.product_error_tolerance
        peptide_masses = self.estimate_peptide_mass(
            scan, mass_shift=mass_shift, query_mass=query_mass)
        peptide_masses = [PPMQueryInterval(
            p, error_tolerance) for p in peptide_masses]
        if self.use_denovo_motif:
            path_masses = self.motif_finder.build_peptide_filter(
                scan, error_tolerance, mass_shift=mass_shift)
            peptide_masses.extend(path_masses)
        peptide_masses.sort(key=lambda x: x.center)

        if len(peptide_masses) == 0:
            return IntervalFilter([])
        out = IntervalFilter(peptide_masses)
        out.compress()
        return out


class GlycanFilteringPeptideMassEstimator(
        GlycanFilteringPeptideMassEstimatorBase[GlycanMatchResult], GlycanCoarseScorerBase):
    def n_glycan_coarse_score(self, scan: ProcessedScan,
                              glycan_combination: GlycanCombinationRecord,
                              mass_shift: MassShift = Unmodified,
                              peptide_mass: Optional[float] = None) -> Tuple[float, CoarseGlycanMatch]:
        '''Calculates a ranking score from N-glycopeptide stub-glycopeptide fragments.

        This method is derived from the technique used in pGlyco2 [1].

        References
        ----------
        [1] Liu, M.-Q., Zeng, W.-F., Fang, P., Cao, W.-Q., Liu, C., Yan, G.-Q., … Yang, P.-Y. (2017).
            pGlyco 2.0 enables precision N-glycoproteomics with comprehensive quality control and
            one-step mass spectrometry for intact glycopeptide identification. Nature Communications,
            8(1), 438. https://doi.org/10.1038/s41467-017-00535-2
        '''
        if peptide_mass is None:
            peptide_mass = (
                scan.precursor_information.neutral_mass - glycan_combination.dehydrated_mass
            ) - mass_shift.mass
        if peptide_mass < 0:
            return -1e6, None
        glycan_match = self._n_glycan_match_stubs(
            scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=mass_shift.tandem_mass)
        score = self._calculate_score(glycan_match)
        return score, glycan_match

    def o_glycan_coarse_score(self, scan: ProcessedScan,
                              glycan_combination: GlycanCombinationRecord,
                              mass_shift: MassShift = Unmodified,
                              peptide_mass: Optional[float] = None) -> Tuple[float, CoarseGlycanMatch]:
        '''Calculates a ranking score from O-glycopeptide stub-glycopeptide fragments.

        This method is derived from the technique used in pGlyco2 [1].

        References
        ----------
        [1] Liu, M.-Q., Zeng, W.-F., Fang, P., Cao, W.-Q., Liu, C., Yan, G.-Q., … Yang, P.-Y. (2017).
            pGlyco 2.0 enables precision N-glycoproteomics with comprehensive quality control and
            one-step mass spectrometry for intact glycopeptide identification. Nature Communications,
            8(1), 438. https://doi.org/10.1038/s41467-017-00535-2
        '''
        if peptide_mass is None:
            peptide_mass = (
                scan.precursor_information.neutral_mass - glycan_combination.dehydrated_mass
            ) - mass_shift.mass
        if peptide_mass < 0:
            return -1e6, None
        glycan_match = self._o_glycan_match_stubs(
            scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=mass_shift.tandem_mass)
        score = self._calculate_score(glycan_match)
        return score, glycan_match

    def gag_coarse_score(self, scan: ProcessedScan,
                         glycan_combination: GlycanCombinationRecord,
                         mass_shift: MassShift = Unmodified,
                         peptide_mass: Optional[float] = None) -> Tuple[float, CoarseGlycanMatch]:
        '''Calculates a ranking score from GAG linker glycopeptide stub-glycopeptide fragments.

        This method is derived from the technique used in pGlyco2 [1].

        References
        ----------
        [1] Liu, M.-Q., Zeng, W.-F., Fang, P., Cao, W.-Q., Liu, C., Yan, G.-Q., … Yang, P.-Y. (2017).
            pGlyco 2.0 enables precision N-glycoproteomics with comprehensive quality control and
            one-step mass spectrometry for intact glycopeptide identification. Nature Communications,
            8(1), 438. https://doi.org/10.1038/s41467-017-00535-2
        '''
        if peptide_mass is None:
            peptide_mass = (
                scan.precursor_information.neutral_mass - glycan_combination.dehydrated_mass
            ) - mass_shift.mass
        if peptide_mass < 0:
            return -1e6, None
        glycan_match = self._gag_match_stubs(
            scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=mass_shift.tandem_mass)
        score = self._calculate_score(glycan_match)
        return score, glycan_match

    def match(self, scan, mass_shift=Unmodified, query_mass=None) -> List[GlycanMatchResult]:
        output = []
        if query_mass is None:
            intact_mass = scan.precursor_information.neutral_mass
        else:
            intact_mass = query_mass
        threshold_mass = (intact_mass + 1) - self.minimum_peptide_mass
        for glycan_combination in self.glycan_combination_db:
            # Stop searching when the peptide mass would be below the minimum peptide mass
            if threshold_mass < glycan_combination.dehydrated_mass:
                break
            peptide_mass = (
                intact_mass - glycan_combination.dehydrated_mass
            ) - mass_shift.mass
            best_score = 0
            best_match = None
            type_to_score = {}
            if glycan_combination.is_n_glycan():
                score, match = self.n_glycan_coarse_score(
                    scan, glycan_combination, mass_shift=mass_shift, peptide_mass=peptide_mass)
                type_to_score[GlycanTypes.n_glycan] = (score, match)
                if score > best_score:
                    best_score = score
                    best_match = match
            if glycan_combination.is_o_glycan():
                score, match = self.o_glycan_coarse_score(
                    scan, glycan_combination, mass_shift=mass_shift, peptide_mass=peptide_mass)
                type_to_score[GlycanTypes.o_glycan] = (score, match)
                if score > best_score:
                    best_score = score
                    best_match = match
            if glycan_combination.is_gag_linker():
                score, match = self.gag_coarse_score(
                    scan, glycan_combination, mass_shift=mass_shift, peptide_mass=peptide_mass)
                type_to_score[GlycanTypes.gag_linker] = (score, match)
                if score > best_score:
                    best_score = score
                    best_match = match

            if best_match is not None:
                recalibrated_peptide_mass = best_match.estimate_peptide_mass()
                if recalibrated_peptide_mass > 0:
                    if abs(recalibrated_peptide_mass - peptide_mass) > 0.5:
                        warnings.warn("Re-estimated peptide mass error is large: %f vs %f" % (
                            peptide_mass, recalibrated_peptide_mass))
            else:
                recalibrated_peptide_mass = 0
            result = GlycanMatchResult(
                peptide_mass,
                best_score, best_match, glycan_combination.size, type_to_score, recalibrated_peptide_mass)
            output.append(result)
        output = sorted(output, key=lambda x: x.score, reverse=1)
        return output


class IntervalFilter(Sequence):
    def __init__(self, intervals):
        self.intervals = intervals

    def test(self, mass):
        for i in self.intervals:
            if mass in i:
                return True
        return False

    def __getitem__(self, i):
        return self.intervals[i]

    def __len__(self):
        return len(self.intervals)

    def __call__(self, mass):
        return self.test(mass)

    def compress(self):
        if len(self) == 0:
            return self
        out = []
        last = self[0]
        for interval in self[1:]:
            if interval.overlaps(last):
                last.extend(interval)
            else:
                out.append(last)
                last = interval
        out.append(last)
        self.intervals = out
        return self


try:
    has_c = True
    _IntervalFilter = IntervalFilter
    _CoarseStubGlycopeptideFragment = CoarseStubGlycopeptideFragment
    _CoarseGlycanMatch = CoarseGlycanMatch
    from glycresoft._c.structure.intervals import IntervalFilter
    from glycresoft._c.tandem.core_search import (
        CoarseStubGlycopeptideFragment, CoarseGlycanMatch, GlycanMatchResult,
        GlycanFilteringPeptideMassEstimator_match)

    GlycanFilteringPeptideMassEstimator.match = GlycanFilteringPeptideMassEstimator_match
except ImportError:
    has_c = False


class IndexGlycanCompositionFragment(GlycanCompositionFragment):
    __slots__ = ('index', 'name')

    index: int
    name: str

    def __init__(self, mass, composition, key, is_extended=False):
        self.mass = mass
        self.composition = composition
        self.key = key
        self.is_extended = is_extended
        self._hash_key = -1
        self.index = -1
        self.name = None

    def __reduce__(self):
        return self.__class__, (self.mass, self.composition, self.key, self.is_extended), self.__getstate__()

    def __getstate__(self):
        return {
            "index": self.index,
            "name": self.name,
            "_hash_key": self._hash_key
        }

    def __setstate__(self, state):
        self.index = state['index']
        self.name = state['name']
        self._hash_key = state['_hash_key']


class ComplementFragment(object):
    __slots__ = ('mass', 'keys')

    mass: float
    keys: List[Tuple[IndexGlycanCompositionFragment, int, Any]]

    def __init__(self, mass, keys=None):
        self.mass = mass
        self.keys = keys or []

    def __repr__(self):
        return "{self.__class__.__name__}({self.mass})".format(self=self)


GlycanTypes_n_glycan = GlycanTypes.n_glycan
GlycanTypes_o_glycan = GlycanTypes.o_glycan
GlycanTypes_gag_linker = GlycanTypes.gag_linker


class PartialGlycanSolution(object):
    __slots__ = ("peptide_mass", "score", "core_matches", "fragment_matches", "glycan_index")

    peptide_mass: float
    score: float
    core_matches: Set[int]
    fragment_matches: Set[int]
    glycan_index: int

    def __init__(self, peptide_mass=-1, score=0, core_matches=None, fragment_matches=None, glycan_index=-1):
        if core_matches is None:
            core_matches = set()
        if fragment_matches is None:
            fragment_matches = set()
        self.peptide_mass = peptide_mass
        self.score = score
        self.core_matches = core_matches
        self.fragment_matches = fragment_matches
        self.glycan_index = glycan_index

    def __repr__(self):
        template = ("{self.__class__.__name__}({self.peptide_mass}, {self.score}, "
                    "{self.core_matches}, {self.fragment_matches}, {self.glycan_index})")
        return template.format(self=self)

    @property
    def fragment_match_count(self):
        return len(self.fragment_matches)


class GlycanFragmentIndex(object):
    '''A fast and sparse in-memory fragment ion index for quickly matching peaks against multiple
    glycan composition complements.

    Based upon the complement ion indexing strategy described in [1]_.

    Attributes
    ----------
    members : :class:`list` of :class:`GlycanCombinationRecord`
        The glycan composition combinations in this index.
    unique_fragments : :class:`dict`[:class:`str`, :class:`dict`[:class:`str`, :class:`IndexGlycanCompositionFragment`]]
        An internment table for each glycosylation type mapping fragment name to glycan composition fragments
    fragment_index : :class:`dict`[:class:`int`, :class:`list`[:class:`ComplementFragment`]]
        A sparse index mapping :attr:`resolution` scaled masses to bins. The mass values and binned
        fragments are complements of true fragments.
    counter : int
        A counter to assign each unique fragment a unique integer index.
    fragment_weight : float
        A scoring parameter to weight overall coverage with.
    core_weight : float
        A scoring parameter to weight core motif coverage with.
    resolution : float
        A scaling factor to convert real masses into truncated bin indices.


    References
    ----------
    ..[1] Zeng, W., Cao, W., Liu, M., He, S., & Yang, P. (2021). Precise, Fast and
      Comprehensive Analysis of Intact Glycopeptides and Monosaccharide-Modifications with pGlyco3.
      Bioarxiv. https://doi.org/https://doi.org/10.1101/2021.02.06.430063
    '''

    members: List[GlycanCombinationRecord]
    unique_fragments: DefaultDict[str, Dict[str, IndexGlycanCompositionFragment]]
    fragment_index: DefaultDict[int, List[ComplementFragment]]
    counter: int

    core_weight: float
    fragment_weight: float

    resolution: float
    lower_bound: float
    upper_bound: float


    def __init__(self, members=None, fragment_weight=0.56, core_weight=0.42, resolution=100):
        self.members = members or []
        self.unique_fragments = defaultdict(dict)
        self._fragments = []
        self.fragment_index = defaultdict(list)
        self.counter = 0
        self.fragment_weight = fragment_weight
        self.core_weight = core_weight
        self.lower_bound = float('inf')
        self.upper_bound = 0
        self.resolution = resolution

    def _intern(self, fragment: StubFragment, glycosylation_type: str) -> IndexGlycanCompositionFragment:
        key = str(HashableGlycanComposition(fragment.key))
        try:
            return self.unique_fragments[glycosylation_type][key]
        except KeyError:
            fragment = IndexGlycanCompositionFragment(
                fragment.mass, None, fragment.key, not fragment.is_core)
            self.unique_fragments[glycosylation_type][key] = fragment
            fragment.name = key
            fragment.index = self.counter
            self._fragments.append(fragment)
            self.counter += 1
            return fragment

    def _get_fragments(self, gcr: GlycanCombinationRecord, glycosylation_type: str) -> List[StubFragment]:
        if glycosylation_type == GlycanTypes_n_glycan:
            return gcr.get_n_glycan_fragments()
        elif glycosylation_type == GlycanTypes_o_glycan:
            return gcr.get_o_glycan_fragments()
        elif glycosylation_type == GlycanTypes_gag_linker:
            return gcr.get_gag_linker_glycan_fragments()
        else:
            raise ValueError(glycosylation_type)

    def build(self):
        self.members.sort(key=lambda x: (x.dehydrated_mass, x.id))
        self.fragment_index.clear()
        j = 0
        low = float('inf')
        high = 0
        for i, member in enumerate(self.members):
            mass = member.dehydrated_mass
            n_core = 0
            n_frags = 0
            for glycosylation_type in member.glycan_types:
                for frag in self._get_fragments(member, glycosylation_type):
                    frag = self._intern(frag, glycosylation_type)
                    if not frag.is_extended:
                        n_core += 1
                    n_frags += 1
                    d = mass - frag.mass
                    if d < 0:
                        d = 0
                    if high < d:
                        high = d
                    if low > d:
                        low = d
                    key = int(d * self.resolution)
                    for comp in self.fragment_index[key]:
                        if abs(comp.mass - d) <= 1e-3:
                            comp.keys.append((frag, i, glycosylation_type))
                            break
                    else:
                        self.fragment_index[key].append(ComplementFragment(d, [(frag, i, glycosylation_type)]))
                        j += 1

                member.fragment_set_properties[glycosylation_type] = (n_core, n_frags)
            member.clear()
        self.lower_bound = low
        self.upper_bound = high
        return j

    def _match_fragments(self, delta_mass: float, peak: DeconvolutedPeak,
                         error_tolerance: float, result: Dict[int, Dict[str, PartialGlycanSolution]]):
        key = int(delta_mass * self.resolution)
        width = int(peak.neutral_mass * error_tolerance * self.resolution) + 1
        for off in range(-width, width + 1):
            for comp_frag in self.fragment_index[key + off]:
                if abs(comp_frag.mass - delta_mass) / peak.neutral_mass <= error_tolerance:
                    for frag, i, glycosylation_type in comp_frag.keys:
                        sol = result[i][glycosylation_type]
                        sol.score += np.log10(peak.intensity) * (
                            1 - ((abs(delta_mass - comp_frag.mass) / peak.neutral_mass) / error_tolerance) ** 4)
                        if not frag.is_extended:
                            sol.core_matches.add(frag.index)
                        sol.fragment_matches.add(frag.index)

    def match(self, scan: ProcessedScan, error_tolerance: float=1e-5,
              mass_shift: MassShift=Unmodified, query_mass: Optional[float]=None) -> List[PartialGlycanSolution]:
        if query_mass is None:
            precursor_mass: float = scan.precursor_information.neutral_mass
        else:
            precursor_mass = query_mass

        mass_shift_mass = mass_shift.mass
        mass_shift_tandem_mass = mass_shift.tandem_mass

        result: DefaultDict[int, DefaultDict[str, PartialGlycanSolution]] = defaultdict(
            lambda: defaultdict(PartialGlycanSolution))
        for peak in scan.deconvoluted_peak_set:
            d = (precursor_mass - peak.neutral_mass - mass_shift_mass)
            self._match_fragments(d, peak, error_tolerance, result)
            if mass_shift_tandem_mass != 0:
                d = (precursor_mass - peak.neutral_mass - mass_shift_mass + mass_shift_tandem_mass)
                self._match_fragments(d, peak, error_tolerance, result)

        out = []
        for i, glycosylation_type_to_solutions in result.items():
            rec = self.members[i]
            best_score = -float('inf')
            best_solution = None
            for glycosylation_type, sol in glycosylation_type_to_solutions.items():
                n_core, n_frag = rec.fragment_set_properties[glycosylation_type]
                coverage = (len(sol.core_matches) * 1.0 / n_core) ** self.core_weight * (
                    len(sol.fragment_matches) * 1.0 / n_frag) ** self.fragment_weight
                sol.score *= coverage
                sol.peptide_mass = precursor_mass - rec.dehydrated_mass - mass_shift_mass
                sol.glycan_index = i
                if best_score < sol.score and sol.peptide_mass > 0:
                    best_score = sol.score
                    best_solution = sol
            if best_solution is not None:
                out.append(sol)

        out.sort(key=lambda x: x.score, reverse=True)
        return out


try:
    _GlycanFragmentIndex = GlycanFragmentIndex
    _IndexGlycanCompositionFragment = IndexGlycanCompositionFragment
    _ComplementFragment = ComplementFragment
    _PartialGlycanSolution = PartialGlycanSolution
    from glycresoft._c.tandem.core_search import (
        GlycanFragmentIndex,
        IndexGlycanCompositionFragment, ComplementFragment, PartialGlycanSolution)
except ImportError:
    pass


class _adapt(object):
    def __init__(self, *args, **kwargs):
        pass


class IndexedGlycanFilteringPeptideMassEstimator(GlycanFilteringPeptideMassEstimatorBase[PartialGlycanSolution],
                                                 _adapt):
    """
    An inverted complement search index for quickly searching for estimating peptide backbone masses
    and glycans to go with them.

    Examples
    --------

    .. code-block:: python

        from glycresoft.tandem.glycopeptide.core_search import (
            GlycanCombinationRecord, IndexedGlycanFilteringPeptideMassEstimator)
        from glycresoft import serialize

        def make_glycan_index(db_path, glycan_hypothesis_id, product_error_tolerance: float=2e-5):
            # Or any object that has a `session` attribute could be used instead of making one here
            db_conn = serialize.DatabaseBoundOperation(db_path)
            # This is the set of *all* glycan combinations for that hypothesis.
            glycan_combinations = GlycanCombinationRecord.from_hypothesis(db_conn, glycan_hypothesis_id)
            search_index = IndexedGlycanFilteringPeptideMassEstimator(
                glycan_combinations, product_error_tolerance=product_error_tolerance)
            return search_index

        db_path = ... # magically get a database path
        scan = ... # magically get a `ProcessedScan` to demonstrate on

        search_index = make_glycan_index(db_path, 1)
        peptide_backbone_mass_candidates = search_index.match(scan)
        for rec in peptide_backbone_mass_candidates:
            candidate_glycans = search_index.glycan_for_peptide_mass(scan, rec.peptide_mass)
            print(rec, candidate_glycans)
    """
    product_error_tolerance: float
    fragment_weight: float
    core_weight: float

    def __init__(self, glycan_combination_db, product_error_tolerance=1e-5,
                 fragment_weight=0.56, core_weight=0.42, minimum_peptide_mass=500.0,
                 use_denovo_motif=False, components=None,
                 use_recalibrated_peptide_mass=False):
        super(IndexedGlycanFilteringPeptideMassEstimator, self).__init__(
            glycan_combination_db, product_error_tolerance, fragment_weight, core_weight,
            minimum_peptide_mass, use_denovo_motif, components, use_recalibrated_peptide_mass)
        self.product_error_tolerance = product_error_tolerance
        self.fragment_weight = fragment_weight
        self.core_weight = core_weight
        self.index = GlycanFragmentIndex(
            self.glycan_combination_db, fragment_weight=fragment_weight, core_weight=core_weight)
        self.index.build()

    def match(self, scan: ProcessedScan, mass_shift: MassShift=Unmodified,
              query_mass: Optional[float]=None) -> List[PartialGlycanSolution]:
        return self.index.match(
            scan,
            error_tolerance=self.product_error_tolerance,
            mass_shift=mass_shift,
            query_mass=query_mass)



import os

from typing import Any, Hashable, List, Mapping, Tuple, Optional, TypedDict, Union, Deque

from glycresoft.task import TaskBase
from glycresoft.chromatogram_tree.mass_shift import MassShift


debug_mode = bool(os.environ.get("GLYCRESOFTDEBUG"))

HitID = Hashable
GroupID = Hashable
DBHit = Any
WorkItem = Tuple[DBHit, List[Tuple[str, MassShift]]]


class WorkItemGroup(TypedDict):
    work_orders: Mapping[HitID, WorkItem]


class StructureSpectrumSpecificationBuilder(object):
    """Base class for building structure hit by spectrum specification
    """

    def build_work_order(self, hit_id: HitID, hit_map: Mapping[HitID, DBHit],
                         scan_hit_type_map: Mapping[Tuple[str, HitID], MassShift],
                         hit_to_scan: Mapping[HitID, List[str]]) -> WorkItem:
        """Packs several task-defining data structures into a simple to unpack payload for
        sending over IPC to worker processes.

        Parameters
        ----------
        hit_id : int
            The id number of a hit structure
        hit_map : dict
            Maps hit_id to hit structure
        hit_to_scan : dict
            Maps hit id to list of scan ids
        scan_hit_type_map : dict
            Maps (scan id, hit id) to the type of mass shift
            applied for this match

        Returns
        -------
        tuple
            Packaged message payload
        """
        return (hit_map[hit_id],
                [(s, scan_hit_type_map[s, hit_id])
                 for s in hit_to_scan[hit_id]])


class TaskSourceBase(StructureSpectrumSpecificationBuilder, TaskBase):
    """A base class for building a stream of work items through
    :class:`StructureSpectrumSpecificationBuilder`.
    """

    batch_size = 10000

    def add(self, item: Union[WorkItem, WorkItemGroup]):
        """Add ``item`` to the work stream

        Parameters
        ----------
        item : object
            The work item to deal
        """
        raise NotImplementedError()

    def join(self):
        """Checkpoint that may halt the stream generation.
        """
        return

    def feed(self, hit_map: Mapping[HitID, str],
             hit_to_scan: Mapping[HitID, List[str]],
             scan_hit_type_map: Mapping[Tuple[str, HitID], MassShift]):
        """Push tasks onto the input queue feeding the worker
        processes.

        Parameters
        ----------
        hit_map : dict
            Maps hit id to structure
        hit_to_scan : dict
            Maps hit id to list of scan ids
        scan_hit_type_map : dict
            Maps (hit id, scan id) to the type of mass shift
            applied for this match
        """
        i = 0
        n = len(hit_to_scan)
        seen = dict()
        log_interval = 10000
        for hit_id, scan_ids in hit_to_scan.items():
            i += 1
            hit = hit_map[hit_id]
            # This sanity checking is likely unnecessary, and is a hold-over from
            # debugging redundancy in the result queue. For the moment, it is retained
            # to catch "new" bugs.
            # If a hit structure's id doesn't match the id it was looked up with, something
            # may be wrong with the upstream process. Log this event.
            if hit.id != hit_id:
                self.log("Hit %r doesn't match its id %r" % (hit, hit_id))
                if hit_to_scan[hit.id] != scan_ids:
                    self.log("Mismatch leads to different scans! (%d, %d)" % (
                        len(scan_ids), len(hit_to_scan[hit.id])))
            # If a hit structure has been seen multiple times independent of whether or
            # not the expected hit id matches, something may be wrong in the upstream process.
            # Log this event.
            if hit.id in seen:
                self.log("Hit %r already dealt under hit_id %r, now again at %r" % (
                    hit, seen[hit.id], hit_id))
                raise ValueError(
                    "Hit %r already dealt under hit_id %r, now again at %r" % (
                        hit, seen[hit.id], hit_id))
            seen[hit.id] = hit_id
            if i % self.batch_size == 0 and i:
                self.join()
            try:
                work_order = self.build_work_order(hit_id, hit_map, scan_hit_type_map, hit_to_scan)
                # if debug_mode:
                #     self.log("...... Matching %s against %r" % work_order)
                self.add(work_order)
                # Set a long progress update interval because the feeding step is less
                # important than the processing step. Additionally, as the two threads
                # run concurrently, the feeding thread can log a short interval before
                # the entire process has formally logged that it has started.
                if i % log_interval == 0:
                    self.log("...... Dealt %d work items (%0.2f%% Complete)" % (i, i * 100.0 / n))
            except Exception as e:
                self.log("An exception occurred while feeding %r and %d scan ids: %r" % (hit_id, len(scan_ids), e))
        if i > log_interval:
            self.log("...... Finished dealing %d work items" % (i,))
        self.join()
        return

    def feed_groups(self, hit_map: Mapping[HitID, DBHit],
                    hit_to_scan: Mapping[HitID, List[str]],
                    scan_hit_type_map: Mapping[Tuple[HitID, str], MassShift],
                    hit_to_group: Mapping[HitID, GroupID]):
        """Push task groups onto the input queue feeding the worker
        processes.

        Parameters
        ----------
        hit_map : dict
            Maps hit id to structure
        hit_to_scan : dict
            Maps hit id to list of scan ids
        scan_hit_type_map : dict
            Maps (hit id, scan id) to the type of mass shift
            applied for this match
        hit_to_group: dict
            Maps group id to the set of hit ids which are
        """
        i = 0
        j = 0
        seen = dict()
        for group_key, hit_keys in hit_to_group.items():
            hit_group: WorkItemGroup = {
                "work_orders": {}
            }
            i += 1
            for hit_id in hit_keys:
                j += 1
                scan_ids = hit_to_scan[hit_id]
                hit = hit_map[hit_id]
                # This sanity checking is likely unnecessary, and is a hold-over from
                # debugging redundancy in the result queue. For the moment, it is retained
                # to catch "new" bugs.
                # If a hit structure's id doesn't match the id it was looked up with, something
                # may be wrong with the upstream process. Log this event.
                if hit.id != hit_id:
                    self.log("Hit %r doesn't match its id %r" % (hit, hit_id))
                    if hit_to_scan[hit.id] != scan_ids:
                        self.log("Mismatch leads to different scans! (%d, %d)" % (
                            len(scan_ids), len(hit_to_scan[hit.id])))
                # If a hit structure has been seen multiple times independent of whether or
                # not the expected hit id matches, something may be wrong in the upstream process.
                # Log this event.
                if hit.id in seen:
                    self.log("Hit %r already dealt under hit_id %r, now again at %r in group %r" % (
                        hit, seen[hit.id], hit_id, group_key))
                    raise ValueError(
                        "Hit %r already dealt under hit_id %r, now again at %r" % (
                            hit, seen[hit.id], hit_id))
                seen[hit.id] = (hit_id, group_key)
                work_order = self.build_work_order(
                    hit_id, hit_map, scan_hit_type_map, hit_to_scan)
                hit_group['work_orders'][hit_id] = work_order
            self.add(hit_group)
            if i % self.batch_size == 0 and i:
                self.join()
        self.join()
        return

    def __call__(self, hit_map: Mapping[HitID, Any],
                 hit_to_scan: Mapping[HitID, List[str]],
                 scan_hit_type_map: Mapping[Tuple[HitID, str], MassShift],
                 hit_to_group: Optional[Mapping[HitID, Any]]=None):
        if not hit_to_group:
            return self.feed(hit_map, hit_to_scan, scan_hit_type_map)
        else:
            return self.feed_groups(hit_map, hit_to_scan, scan_hit_type_map, hit_to_group)


class TaskDeque(TaskSourceBase):
    """Generate an on-memory buffer of work items

    Attributes
    ----------
    queue : :class:`~.deque`
        The in-memory work queue
    """

    queue: Deque[Union[WorkItem, WorkItemGroup]]

    def __init__(self):
        self.queue = Deque()

    def add(self, item: Union[WorkItemGroup, WorkItem]):
        self.queue.append(item)

    def pop(self):
        return self.queue.popleft()

    def __iter__(self):
        return iter(self.queue)


class TaskQueueFeeder(TaskSourceBase):
    def __init__(self, input_queue, done_event):
        self.input_queue = input_queue
        self.done_event = done_event

    def add(self, item):
        self.input_queue.put(item)

    def join(self):
        return self.input_queue.join()

    def feed(self, hit_map, hit_to_scan, scan_hit_type_map):
        """Push tasks onto the input queue feeding the worker
        processes.

        Parameters
        ----------
        hit_map : dict
            Maps hit id to structure
        hit_to_scan : dict
            Maps hit id to list of scan ids
        scan_hit_type_map : dict
            Maps (hit id, scan id) to the type of mass shift
            applied for this match
        """
        super(TaskQueueFeeder, self).feed(hit_map, hit_to_scan, scan_hit_type_map)
        self.done_event.set()
        return

    def feed_groups(self, hit_map, hit_to_scan, scan_hit_type_map, hit_to_group):
        super(TaskQueueFeeder, self).feed_groups(hit_map, hit_to_scan, scan_hit_type_map, hit_to_group)
        self.done_event.set()
        return



from __future__ import print_function
import os
import logging
import pprint
import time
import traceback
import multiprocessing
import threading

from typing import Any, Generic, List, Optional, TypeVar, Union
from logging.handlers import QueueHandler, QueueListener
from multiprocessing.managers import SyncManager
from datetime import datetime

from queue import Empty
import warnings

import six

from glycresoft.version import version



logger = logging.getLogger("glycresoft.task")

T = TypeVar("T")


def display_version(print_fn):
    msg = "glycresoft: version %s" % version
    print_fn(msg)


def ensure_text(obj):
    if six.PY2:
        return six.text_type(obj)
    else:
        return str(obj)


def fmt_msg(*message):
    return u"%s %s" % (ensure_text(datetime.now().isoformat(' ')), u', '.join(map(ensure_text, message)))


def printer(obj, *message, stacklevel=None):
    print(fmt_msg(*message))


def debug_printer(obj, *message, stacklevel=None):
    if obj.in_debug_mode():
        print(u"DEBUG:" + fmt_msg(*message))


class CallInterval(object):
    """Call a function every `interval` seconds from
    a separate thread.

    Attributes
    ----------
    stopped: threading.Event
        A semaphore lock that controls when to run `call_target`
    call_target: callable
        The thing to call every `interval` seconds
    args: iterable
        Arguments for `call_target`
    interval: number
        Time between calls to `call_target`
    """

    def __init__(self, interval, call_target, *args):
        self.stopped = threading.Event()
        self.interval = interval
        self.call_target = call_target
        self.args = args
        self.thread = threading.Thread(target=self.mainloop)
        self.thread.daemon = True

    def mainloop(self):
        while not self.stopped.wait(self.interval):
            try:
                self.call_target(*self.args)
            except Exception as e:
                logger.exception("An error occurred in %r", self, exc_info=e)

    def start(self):
        self.thread.start()

    def stop(self):
        self.stopped.set()


class IPCLoggingManager:
    queue: multiprocessing.Queue
    listener: QueueListener

    def __init__(self, queue=None, *handlers):
        if queue is None:
            queue = multiprocessing.Queue()
        if not handlers:
            logger = logging.getLogger()
            handlers = logger.handlers

        self.queue = queue
        self.listener = QueueListener(
            queue, *handlers, respect_handler_level=True)
        self.listener.start()

    def sender(self, logger_name="glycresoft"):
        return LoggingHandlerToken(self.queue, logger_name)

    def start(self):
        self.listener.start()

    def stop(self):
        try:
            self.listener.stop()
        except AttributeError:
            pass


class LoggingHandlerToken:
    queue: multiprocessing.Queue
    name: str
    configured: bool

    def __init__(self, queue: multiprocessing.Queue, name: str):
        self.queue = queue
        self.name = name
        self.configured = False

    def get_logger(self) -> logging.Logger:
        logger = logging.getLogger(self.name)
        return logger

    def clear_handlers(self, logger: logging.Logger):
        for handler in list(logger.handlers):
            logger.removeHandler(handler)
        if logger.parent is not None and logger.parent is not logger:
            self.clear_handlers(logger.parent)

    def log(self, *args, **kwargs):
        kwargs.setdefault('stacklevel', 2)
        self.get_logger().info(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        kwargs.setdefault('stacklevel', 3)
        self.log(*args, **kwargs)

    def add_handler(self):
        if self.configured:
            return
        logger = self.get_logger()
        self.clear_handlers(logger)
        handler = QueueHandler(self.queue)
        handler.setLevel(logging.INFO)
        logger.addHandler(handler)
        logger.setLevel(logging.INFO)
        LoggingMixin.log_with_logger(logger)
        TaskBase.log_with_logger(logger)
        self.configured = True

    def __getstate__(self):
        return {
            "queue": self.queue,
            "name": self.name
        }

    def __setstate__(self, state):
        self.queue = state['queue']
        self.name = state['name']
        self.configured = False
        if multiprocessing.current_process().name == "MainProcess":
            return
        self.add_handler()


class MessageSpooler(object):
    """An IPC-based logging helper

    Attributes
    ----------
    halting : bool
        Whether the object is attempting to
        stop, so that the internal thread can
        tell when it should stop and tell other
        objects using it it is trying to stop
    handler : Callable
        A Callable object which can be used to do
        the actual logging
    message_queue : multiprocessing.Queue
        The Inter-Process Communication queue
    thread : threading.Thread
        The internal listener thread that will consume
        message_queue work items
    """
    def __init__(self, handler):
        self.handler = handler
        self.message_queue = multiprocessing.Queue()
        self.halting = False
        self.thread = threading.Thread(target=self.run)
        self.thread.start()

    def run(self):
        while not self.halting:
            try:
                message = self.message_queue.get(True, 2)
                self.handler(*message)
            except Exception:
                continue

    def stop(self):
        self.halting = True
        self.thread.join()

    def sender(self):
        return MessageSender(self.message_queue)


class MessageSender(object):
    """A simple callable for pushing objects into an IPC
    queue.

    Attributes
    ----------
    queue : multiprocessing.Queue
        The Inter-Process Communication queue
    """
    def __init__(self, queue):
        self.queue = queue

    def __call__(self, *message):
        self.send(*message)

    def send(self, *message):
        self.queue.put(message)


def humanize_class_name(name):
    parts = []
    i = 0
    last = 0
    while i < len(name):
        c = name[i]
        if c.isupper() and i != last:
            if i + 1 < len(name):
                if name[i + 1].islower():
                    part = name[last:i]
                    parts.append(part)
                    last = i
        i += 1
    parts.append(name[last:i])
    return ' '.join(parts)


class LoggingMixin(object):
    logger_state = None
    print_fn = printer
    debug_print_fn = debug_printer
    error_print_fn = printer
    warn_print_fn = warnings.warn

    _debug_enabled = None

    @classmethod
    def log_with_logger(cls, logger):
        cls.logger_state = logger
        cls.print_fn = logger.info
        cls.debug_print_fn = logger.debug
        cls.error_print_fn = logger.error
        cls.warn_print_fn = logger.warning

    def instance_log_with_logger(self, logger):
        self.logger_state = logger
        self.print_fn = logger.info
        self.debug_print_fn = logger.debug
        self.error_print_fn = logger.error
        self.warn_print_fn = logger.warning

    @classmethod
    def log_to_stdout(cls):
        cls.logger_state = None
        cls.print_fn = printer
        cls.debug_print_fn = debug_printer
        cls.error_print_fn = printer
        cls.warn_print_fn = warnings.warn

    def log(self, *message):
        self.print_fn(u', '.join(map(ensure_text, message)), stacklevel=2)

    def debug(self, *message):
        self.debug_print_fn(u', '.join(map(ensure_text, message)), stacklevel=2)

    def error(self, *message, **kwargs):
        exception = kwargs.get("exception")
        self.error_print_fn(u', '.join(
            map(ensure_text, message)), stacklevel=2)
        if exception is not None:
            self.error_print_fn(traceback.format_exc())

    def warn(self, *message, **kwargs):
        self.warn_print_fn(u', '.join(map(ensure_text, message)), stacklevel=2)

    def ipc_logger(self, handler=None):
        return IPCLoggingManager()

    def in_debug_mode(self):
        if self._debug_enabled is None:
            logger_state = self.logger_state
            if logger_state is not None:
                self._debug_enabled = logger_state.isEnabledFor("DEBUG")
        return bool(self._debug_enabled)


class TaskBase(LoggingMixin):
    """A base class for a discrete, named step in a pipeline that
    executes in sequence.

    Attributes
    ----------
    debug_print_fn : Callable
        The function called to print debug messages
    display_fields : bool
        Whether to display fields at the start of execution
    end_time : datetime.datetime
        The time when the task ended
    error_print_fn : Callable
        The function called to print error messages
    logger_state : logging.Logger
        The Logger bound to this task
    print_fn : Callable
        The function called to print status messages
    start_time : datetime.datetime
        The time when the task began
    status : str
        The state of the executing task
    """

    status = "new"

    display_fields = True

    _display_name = None

    @property
    def display_name(self):
        if self._display_name is None:
            return humanize_class_name(self.__class__.__name__)
        else:
            return self._display_name

    def in_debug_mode(self):
        if self._debug_enabled is None:
            logger_state = self.logger_state
            if logger_state is not None:
                self._debug_enabled = logger_state.isEnabledFor(logging.DEBUG)
        return bool(self._debug_enabled)

    def _format_fields(self):
        if self.display_fields:
            return '\n' + pprint.pformat(
                {k: v for k, v in self.__dict__.items()
                 if not (k.startswith("_") or v is None)})
        else:
            return ''

    def display_header(self):
        display_version(self.log)

    def try_set_process_name(self, name=None):
        """
        This helper method may be used to try to change a process's name
        in order to make discriminating which role a particular process is
        fulfilling. This uses a third-party utility library that may not behave
        the same way on all platforms, and therefore this is done for convenience
        only.

        Parameters
        ----------
        name : str, optional
            A name to set. If not provided, will check the attribute ``process_name``
            for a non-null value, or else have no effect.
        """
        if name is None:
            name = getattr(self, 'process_name', None)
        if name is None:
            return
        _name_process(name)

    def _begin(self, verbose=True, *args, **kwargs):
        self.on_begin()
        self.start_time = datetime.now()
        self.status = "started"
        if verbose:
            self.log(
                "Begin %s%s" % (
                    self.display_name,
                    self._format_fields()))

    def _end(self, verbose=True, *args, **kwargs):
        self.on_end()
        self.end_time = datetime.now()
        if verbose:
            self.log("End %s" % self.display_name)
            self.log(self.summarize())

    def on_begin(self):
        pass

    def on_end(self):
        pass

    def summarize(self):
        chunks = [
            "Started at %s." % self.start_time,
            "Ended at %s." % self.end_time,
            "Total time elapsed: %s" % (self.end_time - self.start_time),
            "%s completed successfully." % self.__class__.__name__ if self.status == 'completed' else
            "%s failed with error message %r" % (self.__class__.__name__, self.status),
            ''
        ]
        return '\n'.join(chunks)

    def start(self, *args, **kwargs):
        self._begin(*args, **kwargs)
        try:
            out = self.run()
        except (KeyboardInterrupt) as e:
            logger.exception("An error occurred: %r", e, exc_info=e)
            self.status = e
            out = e
            raise e
        else:
            self.status = 'completed'
        self._end(*args, **kwargs)
        return out

    def interact(self, **kwargs):
        from IPython.terminal.embed import InteractiveShellEmbed, load_default_config
        import sys
        config = kwargs.get('config')
        header = kwargs.pop('header', u'')
        compile_flags = kwargs.pop('compile_flags', None)
        if config is None:
            config = load_default_config()
            config.InteractiveShellEmbed = config.TerminalInteractiveShell
            kwargs['config'] = config
        frame = sys._getframe(1)
        shell = InteractiveShellEmbed.instance(
            _init_location_id='%s:%s' % (
                frame.f_code.co_filename, frame.f_lineno), **kwargs)
        shell(header=header, stack_depth=2, compile_flags=compile_flags,
              _call_location_id='%s:%s' % (frame.f_code.co_filename, frame.f_lineno))
        InteractiveShellEmbed.clear_instance()


log_handle = TaskBase()


class TaskExecutionSequence(TaskBase, Generic[T]):
    """A task unit that executes in a separate thread or process."""

    def __call__(self) -> T:
        result = None
        try:
            if self._running_in_process:
                self.log("%s running on PID %r" % (self, multiprocessing.current_process().pid))
            if os.getenv("GLYCRESOFTPROFILING"):
                import cProfile
                profiler = cProfile.Profile()
                result = profiler.runcall(self.run, standalone_mode=False)
                profiler.dump_stats('glycresoft_performance.profile')
            else:
                result = self.run()
            self.debug("%r Done" % self)
        except Exception as err:
            self.error("An error occurred while executing %s" %
                       self, exception=err)
            result = None
            self.set_error_occurred()
            try:
                self.done_event.set()
            except AttributeError:
                pass
        finally:
            return result

    def run(self) -> T:
        raise NotImplementedError()

    def _get_repr_details(self):
        return ''

    _thread: Optional[Union[threading.Thread, multiprocessing.Process]] = None
    _running_in_process: bool = False
    _error_flag: Optional[threading.Event] = None

    def error_occurred(self) -> bool:
        if self._error_flag is None:
            return False
        else:
            return self._error_flag.is_set()

    def set_error_occurred(self):
        if self._error_flag is None:
            return False
        else:
            return self._error_flag.set()

    def __repr__(self):
        template = "{self.__class__.__name__}({details})"
        return template.format(self=self, details=self._get_repr_details())

    def _make_event(self, provider=None) -> Union[threading.Event, multiprocessing.Event]:
        if provider is None:
            provider = threading
        return provider.Event()

    def _name_for_execution_sequence(self):
        return ("%s-%r" % (self.__class__.__name__, id(self)))

    def start(self, process: bool=False, daemon: bool=False):
        if self._thread is not None:
            return self._thread
        if process:
            self._running_in_process = True
            self._error_flag = self._make_event(multiprocessing)
            t = multiprocessing.Process(
                target=self, name=self._name_for_execution_sequence())
            if daemon:
                t.daemon = daemon
        else:
            self._error_flag = self._make_event(threading)
            t = threading.Thread(
                target=self, name=self._name_for_execution_sequence())
            if daemon:
                t.daemon = daemon
        t.start()
        self._thread = t
        return t

    def join(self, timeout: Optional[float]=None) -> bool:
        if self.error_occurred():
            return True
        try:
            return self._thread.join(timeout)
        except KeyboardInterrupt:
            self.set_error_occurred()
            return True

    def is_alive(self):
        if self.error_occurred():
            return False
        return self._thread.is_alive()

    def stop(self):
        if self.is_alive():
            self.set_error_occurred()

    def kill_process(self):
        if self._running_in_process:
            if self.is_alive():
                self._thread.terminate()
        else:
            self.log("Cannot kill a process running in a thread")


class Pipeline(TaskExecutionSequence):
    tasks: List[TaskExecutionSequence]
    error_polling_rate: float

    def __init__(self, tasks, error_polling_rate=1.0):
        self.tasks = tasks
        self.error_polling_rate = error_polling_rate

    def start(self, *args, **kwargs):
        for task in self:
            task.start(*args, **kwargs)

    def join(self, timeout: Optional[float]=None):
        if timeout is not None:
            for task in self:
                task.join(timeout)
        else:
            timeout = self.error_polling_rate
            while True:
                has_error = self.error_occurred()
                if has_error:
                    self.log("... Detected an error flag. Stopping!")
                    self.stop()
                    break
                alive = 0
                for task in self:
                    task.join(0.01)
                    is_alive = task.is_alive()
                    alive += is_alive
                if alive == 0:
                    break
                time.sleep(timeout)

    def is_alive(self):
        alive = 0
        for task in self:
            alive += task.is_alive()
        return alive

    def error_occurred(self) -> int:
        errors = 0
        for task in self.tasks:
            errors += task.error_occurred()
        return errors

    def stop(self):
        for task in self.tasks:
            task.stop()

    def __iter__(self):
        return iter(self.tasks)

    def __len__(self):
        return len(self.tasks)

    def __getitem__(self, i: Union[int, slice]):
        return self.tasks[i]

    def add(self, task):
        self.tasks.append(task)
        return self


class SinkTask(TaskExecutionSequence):
    def __init__(self, in_queue, in_done_event):
        self.in_queue = in_queue
        self.in_done_event = in_done_event
        self.done_event = self._make_event()

    def handle_item(self, task):
        pass

    def process(self):
        has_work = True
        while has_work and not self.error_occurred():
            try:
                item = self.in_queue.get(True, 10)
                self.handle_item(item)
            except Empty:
                if self.in_done_event.is_set():
                    has_work = False
                    break
        self.done_event.set()


def make_shared_memory_manager():
    manager = SyncManager()
    manager.start(_name_process, ("glycresoft-shm", ))
    return manager


def _name_process(name):
    try:
        import setproctitle
        setproctitle.setproctitle(name)
    except (ImportError, AttributeError):
        pass


def elapsed(seconds):
    '''Convert a second count into a human readable duration

    Parameters
    ----------
    seconds : :class:`int`
        The number of seconds elapsed

    Returns
    -------
    :class:`str` :
        A formatted, comma separated list of units of duration in days, hours, minutes, and seconds
    '''
    periods = [
        ('day', 60 * 60 * 24),
        ('hour', 60 * 60),
        ('minute', 60),
        ('second', 1)
    ]

    tokens = []
    for period_name, period_seconds in periods:
        if seconds > period_seconds:
            period_value, seconds = divmod(seconds, period_seconds)
            has_s = 's' if period_value > 1 else ''
            tokens.append("%s %s%s" % (period_value, period_name, has_s))

    return ", ".join(tokens)



import operator

from collections import defaultdict
try:
    from itertools import imap
except ImportError:
    imap = map

import logging


from sqlalchemy import select, join

from glycresoft.serialize import (
    GlycanComposition, Glycopeptide, Peptide,
    func, GlycopeptideHypothesis, GlycanHypothesis,
    DatabaseBoundOperation, GlycanClass, GlycanTypes,
    GlycanCompositionToClass, ProteinSite, Protein, GlycanCombination)

from .mass_collection import SearchableMassCollection, NeutralMassDatabase
from .intervals import (
    PPMQueryInterval, FixedQueryInterval, LRUIntervalSet,
    IntervalSet, MassIntervalNode)
from .index import (ProteinIndex, PeptideIndex, )
from glycresoft.structure.structure_loader import (
    CachingGlycanCompositionParser, CachingGlycopeptideParser,
    CachingPeptideParser)
from .composition_network import CompositionGraph, n_glycan_distance


logger = logging.getLogger("glycresoft.database")

_empty_interval = NeutralMassDatabase([])


# The maximum total number of items across all loaded intervals that a database
# is allowed to hold in memory at once before it must start pruning intervals.
# This is to prevent the program from running out of memory because it loaded
# too many large intervals into memory while not exceeding the maximum number of
# intervals allowed.
DEFAULT_THRESHOLD_CACHE_TOTAL_COUNT = 6e5

# The maximum number of mass intervals to hold in memory at once.
DEFAULT_CACHE_SIZE = 3

# Controls the mass interval loaded from the database when a
# query misses the alreadly loaded intervals. Uses the assumption
# that if a mass is asked for, the next mass asked for will be
# close to it, so by pre-emptively loading more data now, less
# time will be spent searching the disk later. The larger this
# number, the more data that will be preloaded. If this number is
# too large, memory and disk I/O is wasted. If too small, then not
# only does it not help solve the pre-loading problem, but it also
# truncates the range that the in-memory interval search actually covers
# and may lead to missing matches. This means this number must chosen
# carefully given the non-constant error criterion we're using, PPM error.
#
# The default value of 1 is good for 10 ppm error intervals on structures
# up to 1e5 Da
DEFAULT_LOADING_INTERVAL = 1.


class DiskBackedStructureDatabaseBase(SearchableMassCollection, DatabaseBoundOperation):
    """A disk-backed structure database that implements the :class:`~.SearchableMassCollection`
    interface. It includes an in-memory caching system to keep recently used mass intervals in memory
    and optimistically fetches nearby masses around queried masses.

    Attributes
    ----------
    hypothesis_id : int
        The database id number of the hypothesis
    loading_interval : float
        The size of the region around a queried mass (in Daltons) to load when
        optimistically fetching records from the disk.
    cache_size : int
        The maximum number of intervals in-memory mass intervals to cache
    intervals : :class:`~.LRUIntervalSet`
        An LRU caching interval collection over masses read from the disk store
    model_type : type
        The ORM type to query against. It must provide an attribute called ``calculated_mass``
    peptides : :class:`~.PeptideIndex`
        An index for looking up :class:`~.Peptide` ORM objects.
    proteins : :class:`~.ProteinIndex`
        An index for looking up :class:`~.Protein` ORM objects.
    threshold_cache_total_count : int
        The total number of records to retain in the in-memory cache before pruning entries,
        independent of the number of loaded intervals
    """

    def __init__(self, connection, hypothesis_id=1, cache_size=DEFAULT_CACHE_SIZE,
                 loading_interval=DEFAULT_LOADING_INTERVAL,
                 threshold_cache_total_count=DEFAULT_THRESHOLD_CACHE_TOTAL_COUNT,
                 model_type=Glycopeptide):
        DatabaseBoundOperation.__init__(self, connection)
        self.hypothesis_id = hypothesis_id
        self.model_type = model_type
        self.loading_interval = loading_interval
        self.threshold_cache_total_count = threshold_cache_total_count
        self._intervals = LRUIntervalSet([], cache_size)
        self._ignored_intervals = IntervalSet([])
        self.proteins = ProteinIndex(self.session, self.hypothesis_id)
        self.peptides = PeptideIndex(self.session, self.hypothesis_id)

    @property
    def cache_size(self):
        return self._intervals.max_size

    @cache_size.setter
    def cache_size(self, value):
        self._intervals.max_size = value

    @property
    def intervals(self):
        return self._intervals

    def reset(self, **kwargs):
        self.intervals.clear()

    def __reduce__(self):
        return self.__class__, (
            self._original_connection, self.hypothesis_id, self.cache_size,
            self.loading_interval, self.threshold_cache_total_count,
            self.model_type)

    def _make_loading_interval(self, mass, error_tolerance=1e-5):
        width = mass * error_tolerance
        if width > self.loading_interval:
            return FixedQueryInterval(mass, width * 2)
        else:
            return FixedQueryInterval(mass, self.loading_interval)

    def _upkeep_memory_intervals(self):
        """Perform routine maintainence of the interval cache, ensuring its size does not
        exceed the upper limit
        """
        n = len(self._intervals)
        if n > 1:
            while (len(self._intervals) > 1 and
                   self._intervals.total_count >
                   self.threshold_cache_total_count):
                logger.info("Upkeep Memory Intervals %d %d", self._intervals.total_count, len(self._intervals))
                self._intervals.remove_lru_interval()

    def _get_record_properties(self):
        return self.fields

    def _limit_to_hypothesis(self, selectable):
        return selectable.where(Peptide.__table__.c.hypothesis_id == self.hypothesis_id)

    def ignore_interval(self, interval):
        self._ignored_intervals.insert_interval(interval)

    def is_interval_ignored(self, interval):
        is_ignored = self._ignored_intervals.find_interval(interval)
        return is_ignored is not None and is_ignored.contains_interval(interval)

    def insert_interval(self, mass, ppm_error_tolerance=1e-5):
        logger.debug("Calling insert_interval with mass %f", mass)
        node = self.make_memory_interval(mass, ppm_error_tolerance)
        logger.debug("New Node: %r", node)
        # We won't insert this node if it is empty.
        if len(node.group) == 0:
            # Ignore seems to be not-working.
            # self.ignore_interval(FixedQueryInterval(mass))
            return node.group
        nearest_interval = self._intervals.find_interval(node)
        # Should an insert be performed if the query just didn't overlap well
        # with the database?
        if nearest_interval is None:
            # No nearby interval, so we should insert
            logger.debug("No nearby interval for %r", node)
            self._intervals.insert_interval(node)
            return node.group
        elif nearest_interval.overlaps(node):
            logger.debug("Nearest interval %r overlaps this %r", nearest_interval, node)
            nearest_interval = self._intervals.extend_interval(nearest_interval, node)
            return nearest_interval.group
        elif not nearest_interval.contains_interval(node):
            logger.debug("Nearest interval %r didn't contain this %r", nearest_interval, node)
            # Nearby interval didn't contain this interval
            self._intervals.insert_interval(node)
            return node.group
        else:
            # Situation unclear.
            # Not worth inserting, so just return the group
            logger.info("Unknown Condition Overlap %r / %r", node, nearest_interval)
            return nearest_interval.group

    def clear_cache(self):
        self._intervals.clear()

    def has_interval(self, mass, ppm_error_tolerance):
        q = PPMQueryInterval(mass, ppm_error_tolerance)
        match = self._intervals.find_interval(q)
        if match is not None:
            q2 = self._make_loading_interval(mass, ppm_error_tolerance)
            logger.debug("Nearest interval %r", match)
            # We are completely contained in an existing interval, so just
            # use that one.
            if match.contains_interval(q):
                logger.debug("Query interval %r was completely contained in %r", q, match)
                return match.group
            # We overlap with an extending interval, so we should populate
            # the new one and merge them.
            elif match.overlaps(q2):
                q3 = q2.difference(match).scale(1.05)
                logger.debug("Query interval partially overlapped, creating disjoint interval %r", q3)
                match = self._intervals.extend_interval(
                    match,
                    self.make_memory_interval_from_mass_interval(q3.start, q3.end)
                )
                return match.group
            # We might need to insert a new interval
            else:
                logger.debug("Query interval %r did not overlap with %r", q, match)
                if self.is_interval_ignored(q):
                    return match.group
                else:
                    return self.insert_interval(mass, ppm_error_tolerance)
        else:
            logger.debug("No existing interval contained %r", q)
            is_ignored = self._ignored_intervals.find_interval(q)
            if is_ignored is not None and is_ignored.contains_interval(q):
                return _empty_interval
            else:
                return self.insert_interval(mass, ppm_error_tolerance)

    def search_mass_ppm(self, mass, error_tolerance):
        self._upkeep_memory_intervals()
        return self.has_interval(mass, error_tolerance).search_mass_ppm(mass, error_tolerance)

    def search_mass(self, mass, error_tolerance):  # pylint: disable=signature-differs
        return self._search_mass_interval(mass - error_tolerance, mass + error_tolerance)

    def _search_mass_interval(self, start, end):
        model = self.model_type
        return self.session.query(self.model_type).filter(
            model.hypothesis_id == self.hypothesis_id,
            model.calculated_mass.between(
                start, end)).all()

    def on_memory_interval(self, mass, interval):
        if len(interval) > self.threshold_cache_total_count:
            logger.info("Interval Length %d for %f", len(interval), mass)

    def make_memory_interval(self, mass, error_tolerance=None):
        interval = self._make_loading_interval(mass, error_tolerance)
        node = self.make_memory_interval_from_mass_interval(interval.start, interval.end)
        return node

    def make_memory_interval_from_mass_interval(self, start, end):
        logger.debug("Querying the database for masses between %f and %f", start, end)
        out = self._search_mass_interval(start, end)
        logger.debug("Retrieved %d records", len(out))
        self.on_memory_interval((start + end) / 2.0, out)
        mass_db = self._prepare_interval(out)
        # bind the bounds of the returned dataset to the bounds of the query
        node = MassIntervalNode(mass_db)
        node.start = start
        node.end = end
        return node

    def _prepare_interval(self, query_results):
        return NeutralMassDatabase(query_results, operator.attrgetter("calculated_mass"))

    @property
    def lowest_mass(self):
        return self.session.query(func.min(self.model_type.calculated_mass)).filter(
            self.model_type.hypothesis_id == self.hypothesis_id).first()

    @property
    def highest_mass(self):
        return self.session.query(func.max(self.model_type.calculated_mass)).filter(
            self.model_type.hypothesis_id == self.hypothesis_id).first()

    def get_record(self, id):
        return self.session.query(self.model_type).get(id)

    def get_object_by_id(self, id):
        return self.session.query(self.model_type).get(id)

    def get_object_by_reference(self, reference):
        return self.session.query(self.model_type).get(reference.id)

    def search_between(self, lower, upper):
        return self.make_memory_interval_from_mass_interval(lower, upper)

    def valid_glycan_set(self):
        glycan_query = self.query(
            GlycanCombination).join(
            GlycanCombination.components).join(
            GlycanComposition.structure_classes).group_by(GlycanCombination.id)

        glycan_compositions = [
            c.convert() for c in glycan_query.all()]

        return set(glycan_compositions)


class DeclarativeDiskBackedDatabase(DiskBackedStructureDatabaseBase):
    """A base class for creating :class:`DiskBackedStructureDatabaseBase` that map against
    ad hoc SQL queries rather than ORM classes

    Attributes
    ----------
    identity_field: :class:`sqlalchemy.sql.Selectable`
        The database column from which the entity identity should be determined
    mass_field: :class:`sqlalchemy.Column`
        The column to read the mass from
    fields: list of :class:`sqlalchemy.ColumnElement`
        The columns to extract from the database to compose records

    """

    def __init__(self, connection, hypothesis_id=1, cache_size=DEFAULT_CACHE_SIZE,
                 loading_interval=DEFAULT_LOADING_INTERVAL,
                 threshold_cache_total_count=DEFAULT_THRESHOLD_CACHE_TOTAL_COUNT):
        super(DeclarativeDiskBackedDatabase, self).__init__(
            connection, hypothesis_id, cache_size, loading_interval,
            threshold_cache_total_count, None)
        self._glycan_composition_network = None

    def __reduce__(self):
        return self.__class__, (
            self._original_connection, self.hypothesis_id, self.cache_size,
            self.loading_interval, self.threshold_cache_total_count)

    @property
    def glycan_composition_network(self):
        return self._glycan_composition_network

    @property
    def lowest_mass(self):
        q = select(
            [func.min(self.mass_field)]).select_from(
            self.selectable)
        q = self._limit_to_hypothesis(q)
        return self.session.execute(q).scalar()

    @property
    def highest_mass(self):
        q = select(
            [func.max(self.mass_field)]).select_from(
            self.selectable)
        q = self._limit_to_hypothesis(q)
        return self.session.execute(q).scalar()

    def _get_record_properties(self):
        return self.fields

    def _limit_to_hypothesis(self, selectable):
        raise NotImplementedError()

    @property
    def structures(self):
        stmt = self._limit_to_hypothesis(
            select(self._get_record_properties()).select_from(
                self.selectable)).order_by(
            self.mass_field)
        return imap(self._convert, self.session.execute(stmt))

    def __iter__(self):
        stmt = self._limit_to_hypothesis(
            select(self._get_record_properties()).select_from(
                self.selectable)).order_by(
            self.mass_field)
        cursor = self.session.execute(stmt)
        converter = self._convert
        while True:
            block_size = 2 ** 16
            block = cursor.fetchmany(block_size)
            if not block:
                break
            for row in block:
                yield converter(row)

    def __getitem__(self, i):
        stmt = self._limit_to_hypothesis(
            select(self._get_record_properties()).select_from(
                self.selectable)).order_by(
            self.mass_field).offset(i)
        return self._convert(self.session.execute(stmt).fetchone())

    def search_mass(self, mass, error_tolerance):
        return self._search_mass_interval(mass - error_tolerance, mass + error_tolerance)

    def _search_mass_interval(self, start, end):
        conn = self.session.connection()
        stmt = self._limit_to_hypothesis(
            select(self._get_record_properties()).select_from(self.selectable)).where(
            self.mass_field.between(
                start, end))
        return conn.execute(stmt).fetchall()

    def get_record(self, id):
        record = self.session.execute(select(self._get_record_properties()).select_from(
            self.selectable).where(
            self.identity_field == id)).first()
        return record

    def get_all_records(self):
        records = self.session.execute(
            select(
                self._get_record_properties()).select_from(self.selectable))
        return list(records)

    def get_object_by_id(self, id):
        return self._convert(self.get_record(id))

    def get_object_by_reference(self, reference):
        return self._convert(self.get_object_by_id(reference.id))

    def __len__(self):
        try:
            return self.hypothesis.parameters['database_size']
        except KeyError:
            stmt = select([func.count(self.identity_field)]).select_from(
                self.selectable)
            stmt = self._limit_to_hypothesis(stmt)
            return self.session.execute(stmt).scalar()


class GlycopeptideDiskBackedStructureDatabase(DeclarativeDiskBackedDatabase):
    selectable = join(Glycopeptide.__table__, Peptide.__table__)
    fields = [
        Glycopeptide.__table__.c.id,
        Glycopeptide.__table__.c.calculated_mass,
        Glycopeptide.__table__.c.glycopeptide_sequence,
        Glycopeptide.__table__.c.protein_id,
        Peptide.__table__.c.start_position,
        Peptide.__table__.c.end_position,
        Peptide.__table__.c.calculated_mass.label("peptide_mass"),
        Glycopeptide.__table__.c.hypothesis_id,
    ]
    mass_field = Glycopeptide.__table__.c.calculated_mass
    identity_field = Glycopeptide.__table__.c.id

    def __init__(self, connection, hypothesis_id=1, cache_size=DEFAULT_CACHE_SIZE,
                 loading_interval=DEFAULT_LOADING_INTERVAL,
                 threshold_cache_total_count=DEFAULT_THRESHOLD_CACHE_TOTAL_COUNT):
        super(GlycopeptideDiskBackedStructureDatabase, self).__init__(
            connection, hypothesis_id, cache_size, loading_interval,
            threshold_cache_total_count)
        self._convert_cache = CachingGlycopeptideParser()

    def _convert(self, bundle):
        inst = self._convert_cache.parse(bundle)
        inst.hypothesis_id = self.hypothesis_id
        return inst

    @property
    def hypothesis(self):
        return self.session.query(GlycopeptideHypothesis).get(self.hypothesis_id)

    def _limit_to_hypothesis(self, selectable):
        return selectable.where(Glycopeptide.__table__.c.hypothesis_id == self.hypothesis_id)

    def make_key_maker(self):
        return GlycopeptideIdKeyMaker(self, self.hypothesis_id)


class GlycopeptideIdKeyMaker(object):
    def __init__(self, database, hypothesis_id):
        self.database = database
        self.hypothesis_id = hypothesis_id
        self.lookup_map = defaultdict(list)
        for gc in self.database.valid_glycan_set():
            self.lookup_map[str(gc)].append(gc.id)

    def make_id_controlled_structures(self, structure, references):
        glycan_ids = self.lookup_map[structure.glycan_composition]
        result = []
        for glycan_id in glycan_ids:
            for ref in references:
                ref_rec = self.database.query(Glycopeptide).get(ref.id)
                alts = self.database.query(Glycopeptide).filter(
                    Glycopeptide.peptide_id == ref_rec.peptide_id,
                    Glycopeptide.glycan_combination_id == glycan_id).all()
                for alt in alts:
                    result.append(alt.convert())
        return result

    def __call__(self, structure, references):
        return self.make_id_controlled_structures(structure, references)


class InMemoryPeptideStructureDatabase(NeutralMassDatabase):
    def __init__(self, records, source_database=None, sort=True):
        super(InMemoryPeptideStructureDatabase, self).__init__(records, sort=sort)
        self.source_database = source_database

    @property
    def hypothesis(self):
        return self.source_database.hypothesis

    @property
    def hypothesis_id(self):
        return self.source_database.hypothesis_id

    @property
    def session(self):
        return self.source_database.session

    def query(self, *args, **kwargs):
        return self.source_database.query(*args, **kwargs)

    @property
    def peptides(self):
        return self.source_database.peptides

    @property
    def proteins(self):
        return self.source_database.proteins


class GlycanCompositionDiskBackedStructureDatabase(DeclarativeDiskBackedDatabase):
    selectable = (GlycanComposition.__table__)
    fields = [
        GlycanComposition.__table__.c.id, GlycanComposition.__table__.c.calculated_mass,
        GlycanComposition.__table__.c.composition, GlycanComposition.__table__.c.formula,
        GlycanComposition.__table__.c.hypothesis_id
    ]
    mass_field = GlycanComposition.__table__.c.calculated_mass
    identity_field = GlycanComposition.__table__.c.id

    def __init__(self, connection, hypothesis_id=1, cache_size=DEFAULT_CACHE_SIZE,
                 loading_interval=DEFAULT_LOADING_INTERVAL,
                 threshold_cache_total_count=DEFAULT_THRESHOLD_CACHE_TOTAL_COUNT):
        super(GlycanCompositionDiskBackedStructureDatabase, self).__init__(
            connection, hypothesis_id, cache_size, loading_interval,
            threshold_cache_total_count)
        self._convert_cache = CachingGlycanCompositionParser()

    def _convert(self, bundle):
        inst = self._convert_cache(bundle)
        inst.hypothesis_id = self.hypothesis_id
        return inst

    @property
    def glycan_composition_network(self):
        if self._glycan_composition_network is None:
            self._glycan_composition_network = CompositionGraph(tuple(self.structures))
            self._glycan_composition_network.create_edges(1, n_glycan_distance)
        return self._glycan_composition_network

    @property
    def hypothesis(self):
        return self.session.query(GlycanHypothesis).get(self.hypothesis_id)

    def glycan_compositions_of_type(self, glycan_type):
        try:
            glycan_type = glycan_type.name
        except AttributeError:
            glycan_type = str(glycan_type)
        if glycan_type in GlycanTypes:
            glycan_type = GlycanTypes[glycan_type]
        stmt = self._limit_to_hypothesis(
            select(self._get_record_properties()).select_from(
                self.selectable.join(GlycanCompositionToClass).join(GlycanClass)).where(
                GlycanClass.__table__.c.name == glycan_type)).order_by(
            self.mass_field)
        return imap(self._convert, self.session.execute(stmt))

    def glycan_composition_network_from(self, query=None):
        if query is None:
            query = self.structures
        compositions = tuple(query)
        graph = CompositionGraph(compositions)
        graph.create_edges(1)
        return graph

    def _limit_to_hypothesis(self, selectable):
        return selectable.where(GlycanComposition.__table__.c.hypothesis_id == self.hypothesis_id)


class PeptideDiskBackedStructureDatabase(DeclarativeDiskBackedDatabase):
    selectable = Peptide.__table__
    fields = [
        Peptide.__table__.c.id,
        Peptide.__table__.c.calculated_mass,
        Peptide.__table__.c.modified_peptide_sequence,
        Peptide.__table__.c.protein_id,
        Peptide.__table__.c.start_position,
        Peptide.__table__.c.end_position,
        Peptide.__table__.c.hypothesis_id,
        Peptide.__table__.c.n_glycosylation_sites,
        Peptide.__table__.c.o_glycosylation_sites,
        Peptide.__table__.c.gagylation_sites,
    ]
    mass_field = Peptide.__table__.c.calculated_mass
    identity_field = Peptide.__table__.c.id

    def __init__(self, connection, hypothesis_id=1, cache_size=DEFAULT_CACHE_SIZE,
                 loading_interval=DEFAULT_LOADING_INTERVAL,
                 threshold_cache_total_count=int(DEFAULT_THRESHOLD_CACHE_TOTAL_COUNT / 5)):
        super(PeptideDiskBackedStructureDatabase, self).__init__(
            connection, hypothesis_id, cache_size, loading_interval,
            threshold_cache_total_count)
        self._convert_cache = CachingPeptideParser()
        self.peptides = PeptideIndex(self.session, self.hypothesis_id)
        self.proteins = ProteinIndex(self.session, self.hypothesis_id)

    # This should exist, but it conflicts with other conversion mechanisms
    # def _convert(self, bundle):
    #     inst = self._convert_cache.parse(bundle)
    #     inst.hypothesis_id = self.hypothesis_id
    #     return inst

    def _limit_to_hypothesis(self, selectable):
        return selectable.where(Peptide.__table__.c.hypothesis_id == self.hypothesis_id)

    @property
    def hypothesis(self):
        return self.session.query(GlycopeptideHypothesis).get(self.hypothesis_id)

    def spanning_n_glycosylation_site(self):
        q = select(self.fields).select_from(
            self.selectable.join(Protein.__table__).join(ProteinSite.__table__)).where(
                Peptide.spans(ProteinSite.location) & (ProteinSite.name == ProteinSite.N_GLYCOSYLATION)
                & (Protein.hypothesis_id == self.hypothesis_id)).order_by(
            self.mass_field).group_by(Peptide.id)
        return q

    def having_glycosylation_site(self):
        pickle_size = func.length(
            Peptide.__table__.c.o_glycosylation_sites) + \
            func.length(
                Peptide.__table__.c.n_glycosylation_sites)
        min_pickle_size = self.session.query(func.min(pickle_size)).scalar()
        if min_pickle_size is None:
            min_pickle_size = 0
        q = select(self.fields).where(
            (pickle_size > min_pickle_size) &
            (Peptide.__table__.c.hypothesis_id == self.hypothesis_id)
        ).order_by(self.mass_field)
        return q

    def has_protein_sites(self):
        has_sites = self.query(ProteinSite).join(ProteinSite.protein).filter(
            Protein.hypothesis_id == self.hypothesis_id,
            ProteinSite.name == ProteinSite.N_GLYCOSYLATION).first()
        return has_sites is not None




import os
import textwrap

from collections import OrderedDict
from typing import Optional

from glycopeptidepy.structure.glycan import GlycosylationType

from glycresoft.tandem.glycopeptide.dynamic_generation.multipart_fdr import (GlycopeptideFDREstimator)
from glycresoft.tandem.target_decoy import GroupwiseTargetDecoyAnalyzer, TargetDecoyAnalyzer

from glycresoft import serialize
from glycresoft.serialize import (
    Protein, Glycopeptide, IdentifiedGlycopeptide,
    func, MSScan, DatabaseBoundOperation)

from glycresoft.chromatogram_tree import Unmodified
from glycresoft.scoring.elution_time_grouping import ModelEnsemble as RetentionTimeEnsemble

from glycresoft.plotting.glycan_visual_classification import (
    GlycanCompositionClassifierColorizer,
    NGlycanCompositionColorizer)
from glycresoft.plotting import (figax, SmoothingChromatogramArtist)
from glycresoft.plotting.sequence_fragment_logo import glycopeptide_match_logo
from glycresoft.plotting.plot_glycoforms import (
    GlycoformLayout)
from glycresoft.plotting.spectral_annotation import TidySpectrumMatchAnnotator
from glycresoft.tandem.glycopeptide.identified_structure import IdentifiedGlycoprotein
from glycresoft.tandem.glycopeptide.scoring import LogIntensityScorer
from glycresoft.plotting.entity_bar_chart import (
    AggregatedAbundanceArtist, BundledGlycanComposition)
from glycresoft.output.report.base import (
    svguri_plot, png_plot, ReportCreatorBase, svg_plot, svg_optimizer)

from ms_deisotope.data_source import ProcessedRandomAccessScanSource
from ms_deisotope.output import ProcessedMSFileLoader



glycan_colorizer_type_map = {
    GlycosylationType.n_linked: NGlycanCompositionColorizer,
    GlycosylationType.glycosaminoglycan: GlycanCompositionClassifierColorizer({}, 'slateblue'),
    GlycosylationType.o_linked: GlycanCompositionClassifierColorizer({}, 'slateblue')
}


def scale_fix_xml_transform(root):
    view_box_str = root.attrib["viewBox"]
    x_start, y_start, x_end, y_end = map(float, view_box_str.split(" "))
    x_start += 0
    updated_view_box_str = " ".join(map(str, [x_start, y_start, x_end, y_end]))
    root.attrib["viewBox"] = updated_view_box_str
    fig_g = root.find(".//{http://www.w3.org/2000/svg}g[@id=\"figure_1\"]")
    fig_g.attrib["transform"] = "scale(1.0, 1.0)"
    return root


class IdentifiedGlycopeptideDescriberBase(object):
    database_connection: DatabaseBoundOperation
    analysis_id: int
    analysis: serialize.Analysis
    mzml_path: str
    scan_loader: Optional[ProcessedRandomAccessScanSource]
    retention_time_model: Optional[RetentionTimeEnsemble]

    def __init__(self, database_path: str, analysis_id: int, mzml_path: Optional[str]=None):
        self.database_connection = DatabaseBoundOperation(database_path)
        self.analysis_id = analysis_id
        self.analysis = self.session.query(serialize.Analysis).get(self.analysis_id)
        self.mzml_path = mzml_path
        self.scan_loader = None
        self.retention_time_model = None
        self._make_scan_loader()

    def chromatogram_plot(self, glycopeptide: serialize.IdentifiedGlycopeptide):
        ax = figax(dpi=120, figsize=(6, 3.1))
        try:
            SmoothingChromatogramArtist(
                glycopeptide, ax=ax, label_peaks=False,
                colorizer=lambda x: "#48afd0").draw(legend=False, axis_font_size=10)
            ax.set_xlabel("Time (Minutes)", fontsize=10)
            ax.set_ylabel("Relative Abundance", fontsize=10)
            ax.ticklabel_format(axis='y', scilimits=(1, 3))
            return png_plot(ax, bbox_inches='tight', img_width='100%', width=8, dpi=160)
        except ValueError:
            return "<div style='text-align:center;'>No Chromatogram Found</div>"

    def flex_panel(self, content: str):
        wrapper = "<div class='flex-item centered'>{content}</div>"
        return wrapper.format(content=content)

    def spectrum_match_info(self, glycopeptide: serialize.IdentifiedGlycopeptide):
        spectrum_match_ref = glycopeptide.best_spectrum_match
        scan_id = spectrum_match_ref.scan.scan_id
        scan = self.scan_loader.get_scan_by_id(scan_id)
        try:
            mass_shift = spectrum_match_ref.mass_shift
        except Exception:
            mass_shift = Unmodified
        if mass_shift.name != Unmodified.name:
            mass_shift = mass_shift.convert()
        else:
            mass_shift = Unmodified

        match = LogIntensityScorer.evaluate(
            scan,
            glycopeptide.structure.convert(),
            error_tolerance=self.analysis.parameters["fragment_error_tolerance"],
            mass_shift=mass_shift)
        specmatch_artist = TidySpectrumMatchAnnotator(match, ax=figax(dpi=160, figsize=(6, 3)))
        specmatch_artist.draw(fontsize=8, pretty=True)
        specmatch_artist.add_summary_labels()
        annotated_match_ax = specmatch_artist.ax

        scan_title = scan.id
        if len(scan_title) > 60:
            scan_title = '\n'.join(textwrap.wrap(scan_title, 60))

        annotated_match_ax.set_title(scan_title, fontsize=11, y=1.05)
        annotated_match_ax.set_ylabel(
            annotated_match_ax.get_ylabel(), fontsize=10)
        annotated_match_ax.set_xlabel(
            annotated_match_ax.get_xlabel(), fontsize=10)

        sequence_logo_plot = glycopeptide_match_logo(
            match, ax=figax(), annotation_options={"lw": 3, "v_step":0.15, "stroke_slope": 0.1})

        spectrum_plot = svguri_plot(
            annotated_match_ax, svg_width="100%", bbox_inches='tight', height=3 * 1.5,
            optimize=True,
            width=8 * 1.5,
            img_width="100%",
            patchless=True, dpi=120)
        sequence_logo_plot.ax.figure.set_figheight(2.5)
        sequence_logo_plot.ax.figure.set_figwidth(8)
        logo_plot = svg_plot(
            sequence_logo_plot.ax,
            svg_width="100%",
            img_width="100%",
            # xml_transform=scale_fix_xml_transform,
            bbox_inches='tight',
            optimize=False,
            height=1, width=6, patchless=True).decode('utf8')
        match.drop_peaks()
        return dict(
            spectrum_plot=spectrum_plot, logo_plot=logo_plot,
            precursor_mass_accuracy=match.precursor_mass_accuracy(),
            spectrum_match=match)

    def format_rt_prediction_interval(self, glycopeptide: serialize.IdentifiedGlycopeptide):
        iv = self.retention_time_model.predict_interval(glycopeptide, 0.01)
        return self.retention_time_model._truncate_interval(iv)

    def _make_scan_loader(self):
        if self.mzml_path is not None:
            if not os.path.exists(self.mzml_path):
                raise IOError("No such file {}".format(self.mzml_path))
            self.scan_loader = ProcessedMSFileLoader(self.mzml_path)
        else:
            self.mzml_path = self.analysis.parameters['sample_path']
            if not os.path.exists(self.mzml_path):
                raise IOError((
                    "No such file {}. If {} was relocated, you may need to explicily pass the"
                    " corrected file path.").format(
                        self.mzml_path,
                        self.database_connection._original_connection))
            self.scan_loader = ProcessedMSFileLoader(self.mzml_path)

    def draw_glycoforms(self, glycoprotein: IdentifiedGlycoprotein) -> str:
        ax = figax()
        layout = GlycoformLayout(
            glycoprotein, glycoprotein.identified_glycopeptides, ax=ax)
        layout.draw()
        svg = layout.to_svg(scale=2.0, height_padding_scale=1.1)
        svg = svg_optimizer(svg.decode('utf8')).decode('utf8')
        return svg


class IdentifiedGlycopeptideDescriberWorker(IdentifiedGlycopeptideDescriberBase):

    def __call__(self, glycopeptide_id: int):
        glycopeptide = self._glycopeptide_from_id(glycopeptide_id)
        return self.spectrum_match_info(glycopeptide)

    def _glycopeptide_from_id(self, glycopeptide_id: int) -> serialize.IdentifiedGlycopeptide:
        return self.database_connection.query(
            IdentifiedGlycopeptide).get(glycopeptide_id)


class GlycopeptideDatabaseSearchReportCreator(ReportCreatorBase, IdentifiedGlycopeptideDescriberBase):
    threshold: float
    fdr_threshold: float
    use_dynamic_display_mode: int
    hypothesis_id: int

    def __init__(self, database_path, analysis_id, stream=None, threshold=5,
                 fdr_threshold=1,
                 mzml_path=None):
        super(GlycopeptideDatabaseSearchReportCreator, self).__init__(
            database_path, analysis_id, stream)
        self.set_template_loader(os.path.dirname(__file__))
        self.mzml_path = mzml_path
        self.scan_loader = None
        self.threshold = threshold
        self.fdr_threshold = fdr_threshold
        self._total_glycopeptides = None
        self.use_dynamic_display_mode = 0
        self.analysis = self.session.query(serialize.Analysis).get(self.analysis_id)
        self._resolve_hypothesis_id()
        self._build_protein_index()
        self._make_scan_loader()
        self._glycopeptide_counter = 0
        if len(self.protein_index) > 10:
            self.use_dynamic_display_mode = 1
        self.fdr_estimator = self.analysis.parameters.get("fdr_estimator")
        self.retention_time_model = self.analysis.parameters.get("retention_time_model")

    def _spawn(self):
        return IdentifiedGlycopeptideDescriberWorker(self.database_connection, self.analysis_id, self.mzml_path)

    def _resolve_hypothesis_id(self):
        self.hypothesis_id = self.analysis.hypothesis_id
        hypothesis = self.session.query(serialize.GlycopeptideHypothesis).get(self.hypothesis_id)
        if hypothesis is None:
            self.hypothesis_id = 1
            hypothesis = self.session.query(serialize.GlycopeptideHypothesis).get(
                self.hypothesis_id)
            if hypothesis is None:
                raise ValueError("Could not resolve Glycopeptide Hypothesis!")

    def prepare_environment(self):
        super(GlycopeptideDatabaseSearchReportCreator, self).prepare_environment()

    def _build_protein_index(self):
        hypothesis_id = self.hypothesis_id
        theoretical_counts = self.session.query(Protein.name, Protein.id, func.count(Glycopeptide.id)).join(
            Glycopeptide).group_by(Protein.id).filter(
            Protein.hypothesis_id == hypothesis_id).all()
        matched_counts = self.session.query(
            Protein.name, Protein.id, func.count(IdentifiedGlycopeptide.id)).join(Protein.glycopeptides).join(
            IdentifiedGlycopeptide, IdentifiedGlycopeptide.structure_id == Glycopeptide.id).group_by(Protein.id).filter(
            IdentifiedGlycopeptide.ms2_score > self.threshold,
            IdentifiedGlycopeptide.q_value < self.fdr_threshold,
            IdentifiedGlycopeptide.analysis_id == self.analysis_id).all()
        listing = []
        index = {}
        for protein_name, protein_id, glycopeptide_count in theoretical_counts:
            index[protein_id] = {
                "protein_name": protein_name,
                "protein_id": protein_id,
            }
        total_glycopeptides = 0
        for protein_name, protein_id, glycopeptide_count in matched_counts:
            entry = index[protein_id]
            entry['identified_glycopeptide_count'] = glycopeptide_count
            total_glycopeptides += glycopeptide_count
            listing.append(entry)
        self._total_glycopeptides = total_glycopeptides
        self.protein_index = sorted(listing, key=lambda x: x["identified_glycopeptide_count"], reverse=True)
        for protein_entry in self.protein_index:
            protein_entry['protein'] = self.session.query(Protein).get(protein_entry["protein_id"])
        return self.protein_index

    def iterglycoproteins(self):
        n = float(len(self.protein_index))

        for i, row in enumerate(self.protein_index, 1):
            protein = row['protein']
            glycopeptides = self.session.query(
                IdentifiedGlycopeptide).join(Glycopeptide).join(
                Protein).filter(
                    IdentifiedGlycopeptide.analysis_id == self.analysis_id,
                    Glycopeptide.hypothesis_id == self.hypothesis_id,
                    IdentifiedGlycopeptide.ms2_score > self.threshold,
                    IdentifiedGlycopeptide.q_value < self.fdr_threshold,
                    Protein.id == protein.id).all()
            glycoprotein = IdentifiedGlycoprotein(protein, glycopeptides)
            self.status_update(
                "Processing %s (%d/%d) %0.2f%%" % (
                    protein.name, i, n, (i / n * 100)))
            yield i, glycoprotein

    def site_specific_abundance_plots(self, glycoprotein: IdentifiedGlycoprotein):
        axes = OrderedDict()
        for glyco_type in glycoprotein.glycosylation_types:
            for site in sorted(glycoprotein.glycosylation_sites_for(glyco_type)):
                spanning_site = glycoprotein.site_map[glyco_type][site]
                if len(spanning_site) == 0:
                    continue
                bundle = BundledGlycanComposition.aggregate(spanning_site)
                if len(bundle) == 0:
                    continue
                ax = figax()
                AggregatedAbundanceArtist(
                    bundle, ax=ax, colorizer=glycan_colorizer_type_map[glyco_type]).draw()
                ax.set_title("%s Glycans\nat Site %d" % (glyco_type.name, site + 1,), fontsize=18)
                axes[site, glyco_type] = svguri_plot(ax, bbox_inches='tight')
        return axes


    def fdr_estimator_plot(self):
        fdr_estimator = self.fdr_estimator
        if fdr_estimator is None:
            return [self.flex_panel("No FDR estimation data")]
        if isinstance(fdr_estimator, (GroupwiseTargetDecoyAnalyzer, TargetDecoyAnalyzer)):
            ax = figax()
            fdr_estimator.plot(ax=ax)
            svg_figure = svguri_plot(ax.figure)
            return [
                self.flex_panel(svg_figure)
            ]
        if isinstance(fdr_estimator, GlycopeptideFDREstimator):
            ax = figax()
            fdr_estimator.glycan_fdr.plot(ax=ax)
            ax.set_title("Glycan FDR", size=16)
            ax.figure.tight_layout()
            figures = [
                "<div class='flex-item centered'>%s</div>" % svguri_plot(
                    ax.figure)
            ]
            ax = figax()
            fdr_estimator.peptide_fdr.plot(ax=ax)
            ax.set_title("Peptide FDR", size=16)
            ax.figure.tight_layout()
            figures.append("<div class='flex-item centered'>%s</div>" %
                           svguri_plot(ax.figure))
            return figures
        return ["<div class='flex-item'>Could not recognize FDR estimation data</div>"]

    def retention_time_model_plot(self):
        figures = []
        rt_model = self.retention_time_model
        if rt_model is None:
            figures.append(self.flex_panel("No retention time model found"))
        else:
            ax = figax()
            rt_model.plot_factor_coefficients(ax=ax)
            figures.append(self.flex_panel(svg_plot(ax.figure).decode('utf8')))
            ax = figax()
            rt_model.plot_residuals(ax=ax)
            figures.append(self.flex_panel(svg_plot(ax.figure).decode('utf8')))

        return figures

    def track_entry(self, glycopeptide: serialize.IdentifiedGlycopeptide):
        self._glycopeptide_counter += 1
        if self._glycopeptide_counter % 15 == 0:
            self.status_update(
                f" ... {self._glycopeptide_counter}/{self._total_glycopeptides} "
                f"({self._glycopeptide_counter / self._total_glycopeptides * 100:0.2f}%) glycopeptides handled"
            )
        return self._glycopeptide_counter

    def make_template_stream(self):
        template_obj = self.env.get_template("overview.templ")

        ads = serialize.AnalysisDeserializer(
            self.database_connection._original_connection,
            analysis_id=self.analysis_id)

        hypothesis = ads.analysis.hypothesis
        sample_run = ads.analysis.sample_run
        if self.use_dynamic_display_mode:
            self.status_update("Using dynamic display mode")
        template_stream = template_obj.stream(
            analysis=ads.analysis,
            hypothesis=hypothesis,
            sample_run=sample_run,
            protein_index=self.protein_index,
            glycoprotein_iterator=self.iterglycoproteins(),
            renderer=self,
            use_dynamic_display_mode=self.use_dynamic_display_mode)

        return template_stream



import json
import warnings

from collections import defaultdict
from io import StringIO

from glypy.composition import Composition, formula

from glycresoft.chromatogram_tree import Unmodified, MassShift, CompoundMassShift
from .base import (
    MassScalingCountScoringModel,
    UniformCountScoringModelBase,
    ScoringFeatureBase)


class MassShiftScoringModelBase(ScoringFeatureBase):
    feature_type = "mass_shift_score"

    def __init__(self, mass_shift_types=None):
        if mass_shift_types is not None:
            self.mass_shift_types = set(mass_shift_types) | {Unmodified}
        else:
            self.mass_shift_types = None

    def get_state_count(self, chromatogram):
        if self.mass_shift_types is None:
            return len(chromatogram.mass_shifts)
        else:
            return len(
                [t for t in self.mass_shift_types
                 if t in chromatogram.mass_shifts])

    def get_states(self, chromatogram):
        if self.mass_shift_types is None:
            return chromatogram.mass_shifts
        else:
            return [t for t in self.mass_shift_types
                    # if t in chromatogram.mass_shifts
                    if [k for k in chromatogram.mass_shifts if k == t or k.composed_with(t)]
                    ]

    def get_signal_proportions(self, chromatogram):
        fractions = chromatogram.mass_shift_signal_fractions()
        total = sum(fractions.values())
        if self.mass_shift_types is None:
            return {k: v / total for k, v in fractions.items()}
        else:
            proportions = defaultdict(float)
            for t in self.mass_shift_types:
                for k, f in fractions.items():
                    if k.composed_with(t):
                        proportions[t] += f
            total = sum(proportions.values())
            return {k: v / total for k, v in proportions.items()}


class UniformMassShiftScoringModel(MassShiftScoringModelBase, UniformCountScoringModelBase):
    pass


uniform_model = UniformMassShiftScoringModel()


class MassScalingMassShiftScoringModel(MassShiftScoringModelBase, MassScalingCountScoringModel):
    def __init__(self, table, mass_shift_types=None, neighborhood_width=100., fit_information=None):
        MassShiftScoringModelBase.__init__(self, mass_shift_types)
        MassScalingCountScoringModel.__init__(self, table, neighborhood_width)
        self.fit_information = fit_information or {}

    def transform_state(self, state):
        return state.name

    def handle_missing_neighborhood(self, chromatogram, neighborhood, *args, **kwargs):
        warnings.warn(
            ("%f was not found for this mass_shift "
             "scoring model. Defaulting to uniform model") % neighborhood)
        return uniform_model.score(chromatogram, *args, **kwargs)

    def handle_missing_bin(self, chromatogram, bins, key, neighborhood, *args, **kwargs):
        warnings.warn("%s not found for this mass range (%f). Using bin average (%r)" % (
            key, neighborhood, chromatogram.mass_shifts))
        return 0

    def clone(self):
        text_buffer = StringIO()
        self.dump(text_buffer)
        text_buffer.seek(0)
        return self.load(text_buffer)

    def _serialize_mass_shift(self, mass_shift):
        return {"name": mass_shift.name, "composition": formula(mass_shift.composition)}

    def _serialize_compound_mass_shift(self, mass_shift):
        return {
            "name": mass_shift.name,
            "composition": formula(mass_shift.composition),
            "counts": {
                k.name: v for k, v in mass_shift.counts.items()
            },
            "definitions": {
                k.name: formula(k.composition) for k, v in mass_shift.counts.items()
            }
        }

    def _serialize_mass_shift_types(self):
        if self.mass_shift_types is None:
            return None
        mass_shift_types = []
        for mass_shift in self.mass_shift_types:
            if isinstance(mass_shift, CompoundMassShift):
                mass_shift_types.append(self._serialize_compound_mass_shift(mass_shift))
            elif isinstance(mass_shift, MassShift):
                mass_shift_types.append(self._serialize_mass_shift(mass_shift))
        return mass_shift_types

    @classmethod
    def _deserialize_mass_shift(cls, data, memo=None):
        if memo is not None:
            if data['name'] in memo:
                return memo[data['name']]
        inst = MassShift(data['name'], Composition(str(data['composition'])))
        if memo is not None:
            memo[inst.name] = inst
        return inst

    @classmethod
    def _deserialize_compound_mass_shift(cls, data, memo=None):
        if memo is not None:
            if data['name'] in memo:
                return memo[data['name']]
        definitions = {}
        for k, v in data['definitions'].items():
            definitions[k] = cls._deserialize_mass_shift({"name": k, "composition": v})

        components = {}
        for k, v in data['counts'].items():
            components[definitions[k]] = v
        inst = CompoundMassShift(components)

        if memo is not None:
            memo[inst.name] = inst
        return inst

    @classmethod
    def _deserialize_mass_shift_types(cls, data):
        if data is None:
            return None
        mass_shift_types = []
        memo = {
            Unmodified.name: Unmodified
        }
        for entry in data:
            if "counts" in entry:
                mass_shift_types.append(cls._deserialize_compound_mass_shift(entry, memo))
            else:
                mass_shift_types.append(cls._deserialize_mass_shift(entry, memo))
        return mass_shift_types

    def dump(self, file_obj):
        json.dump(
            {
                "neighborhood_width": self.neighborhood_width,
                "table": self.table,
                "mass_shift_types": self._serialize_mass_shift_types(),
                "fit_information": self.fit_information
            },
            file_obj, indent=4, sort_keys=True)

    @classmethod
    def load(cls, file_obj):
        data = json.load(file_obj)
        table = data.pop("table")
        width = float(data.pop("neighborhood_width"))
        mass_shift_types = data.pop("mass_shift_types")
        if mass_shift_types is not None:
            mass_shift_types = cls._deserialize_mass_shift_types(mass_shift_types)
        if isinstance(mass_shift_types, list):
            mass_shift_types = set(mass_shift_types)

        def numeric_keys(table, dtype=float, convert_value=lambda x: x):
            return {abs(dtype(k)): convert_value(v) for k, v in table.items()}

        table = numeric_keys(table)

        return cls(table=table, neighborhood_width=width, mass_shift_types=mass_shift_types)

    @classmethod
    def fit(cls, observations, missing=0.01, mass_shift_types=None, neighborhood_width=100.,
            scale_unmodified=0.25):
        if not scale_unmodified:
            scale_unmodified = 1
        bins = defaultdict(lambda: defaultdict(float))

        self = cls({}, mass_shift_types=mass_shift_types,
                   neighborhood_width=neighborhood_width)

        fit_info = {
            "scale_unmodified": scale_unmodified,
            "missing": missing,
            "track": defaultdict(lambda: defaultdict(list)),
            "count": defaultdict(int)
        }

        all_keys = set()
        for sol in observations:
            neighborhood = self.neighborhood_of(sol.neutral_mass)
            fit_info['count'][neighborhood] += 1
            for c in self.get_states(sol):
                key = self.transform_state(c)
                all_keys.add(key)
                val = sol.total_signal_for(c) / (sol.total_signal)
                bins[neighborhood][key] += val
                fit_info['track'][neighborhood][key].append(val)

        model_table = {}

        all_states = set()
        for level in bins.values():
            all_states.update(level.keys())

        for neighborhood, counts in bins.items():
            for c in all_states:
                if str(c) == Unmodified.name:
                    counts[c] *= scale_unmodified
                counts[str(c)] += missing
            total = sum(counts.values())
            entry = {k: v / total for k, v in counts.items()}
            model_table[neighborhood] = entry

        return cls(model_table, mass_shift_types=mass_shift_types, neighborhood_width=neighborhood_width,
                   fit_information=fit_info)


MassScalingAdductScoringModel = MassScalingMassShiftScoringModel



# Inspired heavily by mokapot
# https://github.com/wfondrie/mokapot
# https://pubs.acs.org/doi/10.1021/acs.jproteome.0c01010


from typing import Any, Dict, List, Optional, Tuple

from array import ArrayType as Array

import numpy as np
from numpy.typing import NDArray

from sklearn.svm import LinearSVC
from sklearn.preprocessing import StandardScaler

from ..spectrum_match import SpectrumMatch, MultiScoreSpectrumMatch

from .base import TargetDecoyAnalyzer, NearestValueLookUp, PeptideScoreTargetDecoyAnalyzer, FDREstimatorBase


def _transform_fast(self: StandardScaler, X: np.ndarray) -> np.ndarray:
    Y = (X - self.mean_)
    Y /= self.scale_
    return  Y


def _fast_decision_function(self: LinearSVC, X: np.ndarray) -> np.ndarray:
    Y = X.dot(self.coef_[0, :])
    Y += self.intercept_
    return Y


class SVMModelBase(FDREstimatorBase):
    dataset: TargetDecoyAnalyzer
    proxy_dataset: TargetDecoyAnalyzer

    scaler: StandardScaler
    model: LinearSVC
    model_args: Dict[str, Any]

    max_iter: int
    train_fdr: float
    worse_than_score: bool
    trained: bool

    def __init__(self, target_matches: Optional[List[MultiScoreSpectrumMatch]]=None,
                 decoy_matches: Optional[List[MultiScoreSpectrumMatch]]=None,
                 train_fdr: float=0.01, max_iter: int=10, **model_args):
        self.scaler = StandardScaler()
        self.train_fdr = train_fdr
        self.max_iter = max_iter
        self.trained = False
        self.model_args = model_args.copy()
        self.model = LinearSVC(dual=False, **self.model_args)
        self.worse_than_score = False
        self.dataset = None
        self.proxy_dataset = None

        if target_matches is not None and decoy_matches is not None:
            tda = self._wrap_dataset(target_matches, decoy_matches)
            _, count = tda.get_count_for_fdr(self.train_fdr)
            if count < 100:
                self.warn(f"Found {count} targets passing {self.train_fdr} FDR threshold, "
                          "too few observations to fit a reliable model")
            self.fit(tda)

    def feature_names(self) ->  List[str]:
        raise NotImplementedError()

    def _wrap_dataset(self, target_matches: List[MultiScoreSpectrumMatch],
                      decoy_matches: List[MultiScoreSpectrumMatch]):
        tda = TargetDecoyAnalyzer(target_matches, decoy_matches, decoy_pseudocount=0.0)
        return tda

    def pack(self):
        self.dataset.pack()
        self.proxy_dataset.pack()

    def __getstate__(self):
        return {
            "scaler": self.scaler,
            "model": self.model,
            "model_args": self.model_args,
            "trained": self.trained,
            "worse_than_score": self.worse_than_score,
            "max_iter": self.max_iter,
            "train_fdr": self.train_fdr,
            "dataset": self.dataset,
            "proxy_dataset": self.proxy_dataset
        }

    def __setstate__(self, state):
        self.scaler = state['scaler']
        self.model = state['model']
        self.model_args = state['model_args']
        self.trained = state['trained']
        self.worse_than_score = state['worse_than_score']
        self.max_iter = state['max_iter']
        self.train_fdr = state['train_fdr']
        self.dataset = state['dataset']
        self.proxy_dataset = state['proxy_dataset']

    def init_model(self):
        return LinearSVC(dual=False, **self.model_args)

    def prepare_model(self, model, features: np.ndarray, labels: List[bool]):
        return model

    def predict(self, X: NDArray[np.floating]) -> NDArray[np.floating]:
        X_ = _transform_fast(self.scaler, X)
        return _fast_decision_function(self.model, X_)

    def _get_psms_labels(
        self, tda: TargetDecoyAnalyzer, fdr_threshold: float = None
    ) -> Tuple[List[MultiScoreSpectrumMatch], NDArray[np.bool_], NDArray[np.bool_]]:
        if fdr_threshold is None:
            fdr_threshold = self.train_fdr
        labels = Array('b')
        is_target = Array('b')
        psms = []
        q_value_map = tda.q_value_map

        for t in tda.targets:
            labels.append(int(q_value_map[tda.get_score(t)] <= fdr_threshold))
            psms.append(t)
            is_target.append(True)

        n_targets = np.frombuffer(labels, dtype=bool).sum()

        for d in tda.decoys:
            labels.append(-1)
            psms.append(d)
            is_target.append(False)

        n_decoys = len(tda.decoys)

        self.log(
            f"Selected {n_targets} target spectra and {n_decoys} decoy spectra selected for training"
        )
        return psms, np.frombuffer(labels, dtype=np.int8), np.frombuffer(is_target, dtype=bool)

    def extract_features(self, psms: List[MultiScoreSpectrumMatch]) -> np.ndarray:
        '''Override in subclass to construct feature matrix for SVM model'''
        raise NotImplementedError()

    def _wrap_psms_with_score(
        self, psms: List[SpectrumMatch], scores: List[float], is_target: List[bool]
    ) -> List[SpectrumMatch]:
        targets = []
        decoys = []
        for psm, score, is_t in zip(psms, scores, is_target):
            wrapped = SpectrumMatch(psm.scan, psm.target, score)
            if is_t:
                targets.append(wrapped)
            else:
                decoys.append(wrapped)
        return targets, decoys

    def scores_to_q_values(
        self, psms: List[SpectrumMatch], scores: np.ndarray, is_target: List[bool]
    ) -> Tuple[NearestValueLookUp, np.ndarray, ]:

        targets, decoys = self._wrap_psms_with_score(psms, scores, is_target)
        tda = TargetDecoyAnalyzer(targets, decoys, decoy_pseudocount=0)
        q_value_map = tda.q_value_map
        q_values = np.array([q_value_map[s] for s in scores])

        updated_labels = (q_values <= self.train_fdr).astype(int)
        updated_labels[~np.asanyarray(is_target)] = -1
        return q_value_map, q_values, updated_labels, tda

    def fit(self, tda: TargetDecoyAnalyzer):
        self.dataset = tda
        psms, labels, is_target = self._get_psms_labels(tda, self.train_fdr)

        k = is_target.sum()
        has_no_decoys = k == len(is_target)
        has_no_targets = k == 0
        if has_no_decoys or has_no_targets:
            self.worse_than_score = True
            self.proxy_dataset = self.dataset
            return

        features = self.extract_features(psms)
        normalized_features = self.scaler.fit_transform(features)
        starting_labels = labels

        model = self.prepare_model(self.model, normalized_features, labels)
        for i in range(self.max_iter):
            observations = normalized_features[labels.astype(bool), :]
            target_labels_i = (labels[labels.astype(bool)] + 1) / 2
            if (target_labels_i == 1).sum() == 0 or (target_labels_i == 0).sum() == 0:
                self.warn("Found only one observation class, cannot proceed with model fitting")
                break
            model.fit(observations, target_labels_i)

            scores = self.model.decision_function(normalized_features)
            _q_value_map, _q_values, labels, self.proxy_dataset = self.scores_to_q_values(
                psms, scores, is_target
            )
            self.log(f"... Round {i}: {(labels == 1).sum()}")

        if (labels == 1).sum() < (starting_labels == 1).sum():
            self.log(
                f"Model performing worse than initial {(labels == 1).sum()} < {(starting_labels == 1).sum()}"
            )
            self.worse_than_score = True
            assert self.dataset is not None
        self.model = model
        self.trained = True
        self._normalized_features = normalized_features

    @property
    def weights(self):
        try:
            return self.model.coef_
        except AttributeError:
            self.log("Model coefficients not yet fit")
            return [[]]

    def plot(self, ax=None):
        if self.worse_than_score:
            return self.dataset.plot(ax=ax)
        ax = self.proxy_dataset.plot(ax=ax)
        lo, hi = ax.get_xlim()
        lo = self.proxy_dataset.thresholds[0] - 0.25
        ax.set_xlim(lo, hi)
        ax.set_xlabel("SVM Score")
        return ax

    def get_count_for_fdr(self, q_value: float):
        return self.proxy_dataset.get_count_for_fdr(q_value)

    @property
    def q_value_map(self):
        return self.proxy_dataset.q_value_map

    @property
    def fdr_map(self):
        return self.q_value_map

    def score(self, spectrum_match: MultiScoreSpectrumMatch, assign=None):
        if assign:
            self.warn("The assign argument is a no-op")
        if self.worse_than_score:
            return self.dataset.score(spectrum_match, assign=False)
        x = self.extract_features([spectrum_match])
        y = self.predict(x)
        return self.q_value_map[y[0]]

    def summarize(self, name: Optional[str]=None):
        if name is None:
            name = "FDR"
        if self.worse_than_score:
            self.dataset.summarize(name)
            return

        self.log("Feature Weights:")
        feature_names = self.feature_names() + ['intercept']
        feature_values = list(self.weights[0]) + [self.model.intercept_]
        for fname, fval in zip(feature_names, feature_values):
            self.log(f"... {fname}: {fval}")

        threshold_05, count_05 = self.get_count_for_fdr(0.05)
        self.log(f"5% {name} = {threshold_05:0.3f} ({count_05})")
        threshold_01, count_01 = self.get_count_for_fdr(0.01)
        self.log(f"1% {name} = {threshold_01:0.3f} ({count_01})")


class PeptideScoreSVMModel(SVMModelBase):

    def _wrap_dataset(self, target_matches: List[MultiScoreSpectrumMatch],
                      decoy_matches: List[MultiScoreSpectrumMatch]):
        tda = PeptideScoreTargetDecoyAnalyzer(
            target_matches, decoy_matches, decoy_pseudocount=0.0)
        return tda

    def feature_names(self) -> List[str]:
        return [
            "peptide_score",
            "peptide_coverage",
        ]

    def extract_features(self, psms: List[MultiScoreSpectrumMatch]) -> np.ndarray:
        features = np.zeros((len(psms), 2))
        for i, psm in enumerate(psms):
            features[i, :] = (
                psm.score_set.peptide_score,
                psm.score_set.peptide_coverage,
            )
        return features




from typing import Type
from sqlalchemy.engine import Connection

from glycresoft.task import log_handle

from .base import Base
from .utils import has_column


class Migration:
    connection: Connection
    tp: Type[Base]

    def __init__(self, connection: Connection, tp: Type[Base]):
        self.connection = connection
        self.tp = tp

    def add_column(self, prop):
        log_handle.log(f"Adding column {prop} for {self.tp}")
        column_name = str(prop.compile(dialect=self.connection.dialect)).split('.')[-1]
        column_type = prop.type.compile(self.connection.dialect)
        migration = prop.info['needs_migration']
        if migration['default'] is None:
            self.connection.execute(
                f'ALTER TABLE {self.tp.__tablename__} ADD COLUMN {column_name} {column_type};')
        else:
            self.connection.execute(
                f'ALTER TABLE {self.tp.__tablename__} ADD COLUMN {column_name} {column_type} DEFAULT {migration["default"]};')

    def needs_migration(self):
        for prop in self.tp.__mapper__.columns:
            if prop.info.get("needs_migration") and not has_column(self.connection, self.tp.__tablename__, prop.name):
                self.add_column(prop)

    def run(self):
        self.needs_migration()



from io import StringIO
from typing import Union, Protocol

from glycopeptidepy import HashableGlycanComposition
from glycopeptidepy.utils import simple_repr

from glycresoft import symbolic_expression


class HasGlycanComposition(Protocol):
    glycan_composition: HashableGlycanComposition


AsGlycanComposition = Union[HashableGlycanComposition, str, HasGlycanComposition]


class CompositionRuleBase(object):

    __repr__ = simple_repr

    def __call__(self, obj: AsGlycanComposition) -> bool:
        raise NotImplementedError()

    def get_composition(self, obj: AsGlycanComposition) -> symbolic_expression.GlycanSymbolContext:
        try:
            composition = obj.glycan_composition
        except AttributeError:
            composition = HashableGlycanComposition.parse(obj)
        composition = symbolic_expression.GlycanSymbolContext(composition)
        return composition

    def __and__(self, other):
        if isinstance(other, CompositionRuleClassifier):
            other = other.copy()
            other.rules.append(self)
            return other
        else:
            return CompositionRuleClassifier(None, [self, other])

    def get_symbols(self):
        raise NotImplementedError()

    @property
    def symbols(self):
        return self.get_symbols()

    def is_univariate(self):
        return len(self.get_symbols()) == 1

    def serialize(self):
        raise NotImplementedError()

    @classmethod
    def parse(cls, line, handle=None):
        raise NotImplementedError()


def int_or_none(x):
    try:
        return int(x)
    except ValueError:
        return None


class CompositionExpressionRule(CompositionRuleBase):
    def __init__(self, expression, required=True):
        self.expression = symbolic_expression.ExpressionNode.parse(str(expression))
        self.required = required

    def get_symbols(self):
        return self.expression.get_symbols()

    def __call__(self, obj):
        composition = self.get_composition(obj)
        if composition.partially_defined(self.expression):
            return composition[self.expression]
        else:
            if self.required:
                return False
            else:
                return True

    def serialize(self):
        tokens = ["CompositionExpressionRule", str(self.expression),
                  str(self.required)]
        return '\t'.join(tokens)

    @classmethod
    def parse(cls, line, handle=None):
        tokens = line.strip().split("\t")
        n = len(tokens)
        i = 0
        while tokens[i] != "CompositionExpressionRule" and i < n:
            i += 1
        i += 1
        if i >= n:
            raise ValueError("Coult not parse %r with %s" % (line, cls))
        expr = symbolic_expression.parse_expression(tokens[i])
        required = tokens[i + 1].lower() in ('true', 'yes', '1')
        return cls(expr, required)

    def __repr__(self):
        template = "{self.__class__.__name__}(expression={self.expression}, required={self.required})"
        return template.format(self=self)


class CompositionRangeRule(CompositionRuleBase):

    def __init__(self, expression, low=None, high=None, required=True):
        self.expression = symbolic_expression.ExpressionNode.parse(str(expression))
        self.low = low
        self.high = high
        self.required = required

    def __repr__(self):
        template = \
            ("{self.__class__.__name__}(expression={self.expression}, "
             "low={self.low}, high={self.high}, required={self.required})")
        return template.format(self=self)

    def get_symbols(self):
        return self.expression.get_symbols()

    def __call__(self, obj: AsGlycanComposition) -> bool:
        composition = self.get_composition(obj)
        if composition.partially_defined(self.expression):
            if self.low is None:
                return composition[self.expression] <= self.high
            elif self.high is None:
                return self.low <= composition[self.expression]
            return self.low <= composition[self.expression] <= self.high
        else:
            if self.required and self.low > 0:
                return False
            else:
                return True

    def serialize(self):
        tokens = ["CompositionRangeRule", str(self.expression), str(self.low),
                  str(self.high), str(self.required)]
        return '\t'.join(tokens)

    @classmethod
    def parse(cls, line, handle=None):
        tokens = line.strip().split("\t")
        n = len(tokens)
        i = 0
        while tokens[i] != "CompositionRangeRule" and i < n:
            i += 1
        i += 1
        if i >= n:
            raise ValueError("Coult not parse %r with %s" % (line, cls))
        expr = symbolic_expression.parse_expression(tokens[i])
        low = int_or_none(tokens[i + 1])
        high = int_or_none(tokens[i + 2])
        required = tokens[i + 3].lower() in ('true', 'yes', '1')
        return cls(expr, low, high, required)


class CompositionRatioRule(CompositionRuleBase):
    def __init__(self, numerator, denominator, ratio_threshold, required=True):
        self.numerator = numerator
        self.denominator = denominator
        self.ratio_threshold = ratio_threshold
        self.required = required

    def __repr__(self):
        template = \
            ("{self.__class__.__name__}(numerator={self.numerator}, "
             "denominator={self.denominator}, ratio_threshold={self.ratio_threshold}, "
             "required={self.required})")
        return template.format(self=self)

    def _test(self, x):
        if isinstance(self.ratio_threshold, (tuple, list)):
            return self.ratio_threshold[0] <= x < self.ratio_threshold[1]
        else:
            return x >= self.ratio_threshold

    def get_symbols(self):
        return (self.numerator, self.denominator)

    def __call__(self, obj: AsGlycanComposition) -> bool:
        composition = self.get_composition(obj)
        val = composition[self.numerator]
        ref = composition[self.denominator]

        if ref == 0 and self.required:
            return False
        else:
            ratio = val / float(ref)
            return self._test(ratio)

    def serialize(self):
        tokens = ["CompositionRatioRule", str(self.numerator), str(self.denominator),
                  str(self.ratio_threshold), str(self.required)]
        return '\t'.join(tokens)

    @classmethod
    def parse(cls, line, handle=None):
        tokens = line.strip().split("\t")
        n = len(tokens)
        i = 0
        while tokens[i] != "CompositionRatioRule" and i < n:
            i += 1
        i += 1
        numerator = symbolic_expression.parse_expression(tokens[i])
        denominator = symbolic_expression.parse_expression(tokens[i + 1])
        ratio_threshold = float(tokens[i + 2])
        required = tokens[i + 3].lower() in ('true', 'yes', '1')
        return cls(numerator, denominator, ratio_threshold, required)


class CompositionRuleClassifier(object):

    def __init__(self, name, rules):
        self.name = name
        self.rules = rules

    def __iter__(self):
        return iter(self.rules)

    def __call__(self, obj: AsGlycanComposition) -> bool:
        for rule in self:
            if not rule(obj):
                return False
        return True

    def __eq__(self, other):
        try:
            return self.name == other.name
        except AttributeError:
            return self.name == other

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.name)

    __repr__ = simple_repr

    def copy(self):
        return CompositionRuleClassifier(self.name, list(self.rules))

    def __and__(self, other):
        if isinstance(other, CompositionRuleClassifier):
            other = other.copy()
            other.rules.extend(self.rules)
            return other
        else:
            self = self.copy()
            self.rules.append(other)
            return self

    def get_symbols(self):
        symbols = set()
        for rule in self:
            symbols.update(rule.symbols)
        return symbols

    @property
    def symbols(self):
        return self.get_symbols()

    def serialize(self):
        text_buffer = StringIO()
        text_buffer.write("CompositionRuleClassifier\t%s\n" % (self.name,))
        for rule in self:
            text = rule.serialize()
            text_buffer.write("\t%s\n" % (text,))
        text_buffer.seek(0)
        return text_buffer.read()

    @classmethod
    def parse(cls, lines):
        line = lines[0]
        name = line.strip().split("\t")[1]
        rules = []
        for line in lines[1:]:
            if line == "":
                continue
            rule_type = line.split("\t")[0]
            if rule_type == "CompositionRangeRule":
                rule = CompositionRangeRule.parse(line)
            elif rule_type == "CompositionRatioRule":
                rule = CompositionRatioRule.parse(line)
            elif rule_type == "CompositionExpressionRule":
                rule = CompositionExpressionRule.parse(line)
            else:
                raise ValueError("Unrecognized Rule Type: %r" % (line,))
            rules.append(rule)
        return cls(name, rules)



import unittest
import tempfile

from glycresoft.database.builder.glycan import glycan_combinator, glycan_source
from glycresoft.database.builder.glycopeptide.common import GlycopeptideHypothesisSerializerBase


class GlycanCombinatoricsTests(unittest.TestCase):
    def setup_tempfile(self):
        file_name = tempfile.mktemp() + '.tmp'
        open(file_name, 'w').write(FILE_SOURCE)
        return file_name

    def clear_file(self, path):
        open(path, 'wb')

    def setup_compositions(self, source_file, database_path):
        builder = glycan_source.TextFileGlycanHypothesisSerializer(source_file, database_path)
        builder.run()
        return builder

    def test_combinator(self):
        source_file = self.setup_tempfile()
        database_path = source_file + '.db'
        builder = self.setup_compositions(source_file, database_path)
        self.assertTrue(builder.query(glycan_combinator.GlycanComposition).count() > 0)

        # Create the glycopeptide hypothesis that GlycanCombination objects must be associated with
        glycopeptide_builder_stub = GlycopeptideHypothesisSerializerBase(database_path, "test", builder.hypothesis_id)

        combinator = glycan_combinator.GlycanCombinationSerializer(
            database_path, builder.hypothesis_id, glycopeptide_builder_stub.hypothesis.id, max_size=2)
        combinator.run()
        inst = builder.query(glycan_combinator.GlycanCombination).filter(
            glycan_combinator.GlycanCombination.hypothesis_id == builder.hypothesis_id,
            glycan_combinator.GlycanCombination.count == 1,
            glycan_combinator.GlycanCombination.composition == "{Hex:5; HexNAc:4; Neu5Ac:2}").one()
        self.assertAlmostEqual(inst.calculated_mass, 2222.7830048, 5)

        inst = builder.query(glycan_combinator.GlycanCombination).filter(
            glycan_combinator.GlycanCombination.hypothesis_id == builder.hypothesis_id,
            glycan_combinator.GlycanCombination.count == 2,
            glycan_combinator.GlycanCombination.composition == "{Hex:10; HexNAc:8; Neu5Ac:3}").first()

        self.assertAlmostEqual(inst.calculated_mass, 4154.47059322789, 5)

        self.clear_file(source_file)
        self.clear_file(source_file + '.db')


FILE_SOURCE = '''
{Hex:5; HexNAc:4; Neu5Ac:1}
{Hex:5; HexNAc:4; Neu5Ac:2}
{Fuc:1; Hex:5; HexNAc:4; Neu5Ac:2}
{Hex:6; HexNAc:4; Neu5Ac:2}
{Fuc:3; Hex:8; HexNAc:4}
{Fuc:2; Hex:6; HexNAc:5; Neu5Ac:1}
{Fuc:1; Hex:6; HexNAc:5; Neu5Ac:2}
{Fuc:3; Hex:6; HexNAc:5; Neu5Ac:1}
{Hex:6; HexNAc:5; Neu5Ac:3}
{Fuc:2; Hex:6; HexNAc:5; Neu5Ac:2}
{Hex:7; HexNAc:6; Neu5Ac:2}
{Fuc:1; Hex:6; HexNAc:5; Neu5Ac:3}
{Fuc:1; Hex:7; HexNAc:6; Neu5Ac:2}
{Hex:10; HexNAc:6; Neu5Ac:1}
{Fuc:2; Hex:6; HexNAc:5; Neu5Ac:3}
{Hex:7; HexNAc:6; Neu5Ac:3}
{Fuc:2; Hex:7; HexNAc:6; Neu5Ac:2}
{Hex:8; HexNAc:7; Neu5Ac:2}
{Fuc:1; Hex:7; HexNAc:6; Neu5Ac:3}
{Hex:7; HexNAc:6; Neu5Ac:4}
{Fuc:2; Hex:7; HexNAc:6; Neu5Ac:3}
{Hex:8; HexNAc:7; Neu5Ac:3}
{Fuc:1; Hex:7; HexNAc:6; Neu5Ac:4}
{Fuc:1; Hex:8; HexNAc:7; Neu5Ac:3}
{Fuc:2; Hex:7; HexNAc:6; Neu5Ac:4}
{Hex:8; HexNAc:7; Neu5Ac:4}
{Fuc:2; Hex:8; HexNAc:7; Neu5Ac:3}
{Fuc:2; Hex:8; HexNAc:10; Neu5Ac:1}
{Fuc:3; Hex:7; HexNAc:6; Neu5Ac:4}
{Hex:9; HexNAc:8; Neu5Ac:3}
{Fuc:5; Hex:7; HexNAc:9; Neu5Ac:1}
{Fuc:1; Hex:8; HexNAc:7; Neu5Ac:4}
{Fuc:2; Hex:8; HexNAc:7; Neu5Ac:4}
{Hex:9; HexNAc:8; Neu5Ac:4}
{Fuc:2; Hex:9; HexNAc:8; Neu5Ac:3}
{Fuc:3; Hex:8; HexNAc:7; Neu5Ac:4}
{Fuc:1; Hex:9; HexNAc:8; Neu5Ac:4}
{Fuc:3; Hex:11; HexNAc:8; Neu5Ac:2}
{Hex:9; HexNAc:8; Neu5Ac:5}
{Fuc:2; Hex:9; HexNAc:8; Neu5Ac:4}
{Fuc:3; Hex:9; HexNAc:8; Neu5Ac:4}
{Fuc:2; Hex:9; HexNAc:8; Neu5Ac:5}
{Fuc:5; Hex:12; HexNAc:9; Neu5Ac:2}
'''

if __name__ == '__main__':
    unittest.main()
